import itertools
import numpy as np

from .. import ParallelSampler


def task(p):
    i,p = p
    results = star_sampler.pipeline.run_results(p)
    #If requested, save the data to file
    if star_sampler.save_name and results.block is not None:
        results.block.save_to_file(star_sampler.save_name+"_%d"%i, clobber=True)
    return (results.post, results.prior, results.extra)

LARGE_JOB_SIZE = 1000000



class StarSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float),("post", float)]
    understands_fast_subspaces = True

    def config(self):
        global star_sampler
        star_sampler = self

        self.converged = False
        self.nsample = self.read_ini("nsample_dimension", int, 1)
        self.save_name = self.read_ini("save", str, "")
        self.nstep = self.read_ini("nstep", int, -1)
        self.allow_large = self.read_ini("allow_large", bool, False)
        self.sample_points = None
        self.ndone = 0

    def setup_sampling(self):
        #Number of jobs to do at once.
        #Can be taken from the ini file.
        #Otherwise it is set to -1 by default
        if self.nstep==-1:
            #if in parallel mode do a chunk of 4*the number of tasks to do at once
            #chosen arbitrarily.
            if self.pool:
                self.nstep = 4*self.pool.size
            #if not parallel then just do a single slice through one dimension each chunk
            else:
                self.nstep = self.nsample

        if self.output:
            for name,value in zip(self.pipeline.varied_params, self.pipeline.start_vector()):
                self.output.metadata("fid_{0}".format(name), value)


        #Also Generate the complete collection of parameter sets to run over.
        #This doesn't actually keep them all in memory, it is just the conceptual
        #outer product
        total_samples = self.nsample*len(self.pipeline.varied_params)
        print()
        print("Total number of star samples: ", total_samples)

        if total_samples>LARGE_JOB_SIZE:
            print("That is a very large number of samples.")
            if self.allow_large:
                print("But you set allow_large=T so I will continue")
            else:
                print("This is suspicously large so I am going to stop")
                print("If you really want to do this set allow_large=T in the")
                print("[star] section of the ini file.")
                raise ValueError("Suspicously large number of star points %d ( = n_samp * n_dim = %d * %d); set allow_large=T in [star] section to permit this."%(total_samples,self.nsample,len(self.pipeline.varied_params)))
        print()
        

        sample_points = []
        start = self.pipeline.start_vector()
        for i,param in enumerate(self.pipeline.varied_params):
            for p in np.linspace(*param.limits, num=self.nsample):
                v = start.copy()
                v[i] = p
                sample_points.append(v)
        self.sample_points = iter(sample_points)




    def execute(self):
        #First run only:
        if self.sample_points is None:
            self.setup_sampling()

        #Chunk of tasks to do this run through, of size nstep.
        #This advances the self.sample_points forward so it knows
        #that these samples have been done
        samples = list(itertools.islice(self.sample_points, self.nstep))

        #If there are no samples left then we are done.
        if not samples:
            self.converged=True
            return

        #Each job has an index number in case we are saving
        #the output results from each one
        sample_index = np.arange(len(samples)) + self.ndone
        jobs = list(zip(sample_index, samples))

        #Actually compute the likelihood results
        if self.pool:
            results = self.pool.map(task, jobs)
        else:
            results = list(map(task, jobs))

        #Update the count
        self.ndone += len(results)

        #Save the results of the sampling
        for sample, result  in zip(samples, results):
            #Optionally save all the results calculated by each
            #pipeline run to files
            (post, prior, extra) = result
            #always save the usual text output
            self.output.parameters(sample, extra, prior, post)

    def is_converged(self):
        return self.converged



#coding: utf-8

u"""Definition of :class:`Pipeline` and the specialization :class:`LikelihoodPipeline`."""


import os
import ctypes
import sys
import string
import numpy as np
import time
import collections
import warnings
import configparser
import traceback
import signal
from . import utils
from . import config
from . import parameter
from . import prior
from . import module
from ..datablock.cosmosis_py import block, section_names
try:
    import faulthandler
    faulthandler.enable()
except ImportError:
    pass


class PipelineResults(object):
    def __init__(self, vector, number_extra):
        self.vector = vector
        self.prior = -np.inf
        self.extra = [np.nan for i in range(number_extra)]
        self.block = None
        self.post = -np.inf
        self.like = -np.inf

    def set_like(self, L):
        self.like = L
        self.post = self.prior + self.like


PIPELINE_INI_SECTION = "pipeline"
NO_LIKELIHOOD_NAMES = "no_likelihood_names_sentinel"

class MissingLikelihoodError(Exception):

    u"""Class to throw if there are not enough data for a likelihood method.

    This class is specifically designed to be thrown during a
    `pipeline.likelihood()` call when there are insufficient data so
    that the test sampler can provide a detailed report to the user
    about the problem (recall that the typical lifetime of a cosmosis
    run involves running the test sampler first to weed out such
    problems, before a full run takes place).

    """

    def __init__(self, message, data):
        u"""Data to provide to the test sampler in case of likelihood data problems.

        The `message` is for the user generally; for use of `data` see
        the `test_sampler.execute()` method, currently at
        `samplers/test/test_sampler.py:29`.
        
        """
        super(MissingLikelihoodError, self).__init__(message)
        self.pipeline_data = data


class LimitedSizeDict(collections.OrderedDict):
    "From http://stackoverflow.com/questions/2437617/limiting-the-size-of-a-python-dictionary"
    def __init__(self, *args, **kwds):
        self.size_limit = kwds.pop("size_limit", None)
        collections.OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key, value):
        collections.OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)


class SlowSubspaceCache(object):
    """
    This tool analyzes pipelines to determine which of their parameters
    are fast and which are slow, and then caches the results of new sets
    of slow parameters so that if only fast parameters have changed the 
    pipeline can be much faster.
    """
    def __init__(self, first_fast_module=None, cache_size_limit=3,):
        self.current_hash = np.nan
        self.analyzed = False
        self.cache = LimitedSizeDict(size_limit=cache_size_limit)
        self.first_fast_module = first_fast_module

    def clear_cache(self):
        self.cache.clear()
        self.current_hash = np.nan

    def hash_slow_parameters(self, block):
        """This is not a general block hash! 
        It just looks at the slow parameters.
        """
        pairs = block.keys()
        pairs.sort()
        values = []
        for section,name in pairs:
            if (section,name) not in self.slow_params:
                continue
            value = block[section, name]
            if not np.isscalar(value):
                #non-scalar in the block - this must
                #reflect some feature not coded yet
                warnings.warn("Vector parameter(s) in the input means that fast-slow split will not work - this can be fixed - please open an issue if it affects you.  Or set fast_slow=F")
                return np.nan # 
            values.append(value)
        values = tuple(values)
        return hash(values)

    def start_pipeline(self, initial_block):
        # We may be in the process of analyzing the pipeline
        # the first time.
        if not self.analyzed:
            return 0
        #
        new_hash = self.hash_slow_parameters(initial_block)
        cached = self.cache.get(new_hash)
        if cached is None:
            self.current_hash = new_hash
            first_module = 0
        else:            
            # Now we need to use the old cached results in the new block.
            # We put everything from the old block into the new block,
            # EXCEPT for the fast parameters themselves.
            for section,name in cached.keys():
                if (section,name) not in self.fast_params:
                    initial_block[section,name] = cached[section,name]
            first_module = self.split_index
        return first_module

    def next_module_results(self, module_index, block):
        if not self.analyzed:
            return

        if module_index != self.split_index-1:
            return

        self.cache[self.current_hash] = block.clone()

    def analyze_pipeline(self, pipeline, all_params=False, grid=False):
        """
        Analyze the pipeline to determine how best to split it into two parts,
        a slow beginning and a faster end, so that we can vary parameters

        """

        #Run the sampler twice first, once to make sure
        #everything is initialized and once to do timing.
        #Use the pipeline starting parameter since that is
        #likely more typical than any random starting position
        start = pipeline.start_vector()

        if pipeline.has_run:
            print("Pipeline has been run once already so no further initialization steps")
        else:
            print("")
            print("Analyzing pipeline to determine fast and slow parameters (because fast_slow=T in [pipeline] section)")
            print("")
            print("Running pipeline once to make sure everything is initialized before timing.")
            print("")
            pipeline.posterior(start)
        print("")
        print("Running the pipeline again to determine timings and fast-slow split")
        print("")
        #Run with timing but make sure to re-set back to the original setting
        #of timing after it finishes
        #This will also print out the timing, which is handy.
        original_timing = pipeline.timing
        pipeline.timing = True
        #Also get the datablock since it contains a log
        #of all the parameter accesses
        _, _, block = pipeline.posterior(start, return_data=True)
        pipeline.timing = original_timing
        timings = pipeline.timings

        if timings is None:
            raise ValueError("Pipeline did not complete, so cannot do fast/slow")

        #Now we have the datablock, which has the log in it, and the timing.
        #The only information that can be of relevance is the fraction
        #of the time pipeline before a given module, and the list of 
        #parameters accessed before that module.
        #We are going to need to go through the log and figure out the latter of
        #these
        if all_params:
            params = pipeline.parameters
        else:
            params = pipeline.varied_params
        first_use = block.get_first_parameter_use(params)
        first_use_count = [len(f) for f in first_use.values()]
        if sum(first_use_count)!=len(params):
            used = sum(first_use.values(), [])
            unused = [p for p in params if p not in used]
            raise ValueError("Tried to do fast-slow split but not all varied parameters ever used in the pipeline: (unused: {})".format(unused))
        print("\n")
        print("Parameters first used in each module:")
        for f, n in zip(first_use.items(), first_use_count):
            name, params = f
            print("{} - {} parameters:".format(name, n))
            for p in params:
                print("     {}--{}".format(*p))

        # Now we have a count of the number of parameters and amount of 
        # time used before each module in the pipeline
        # So we can divide up the parameters into fast and slow.
        self.split_index = self._choose_fast_slow_split(first_use_count, timings, grid)

        self.full_time = sum(timings)
        self.slow_time = sum(timings[:self.split_index])
        self.fast_time = sum(timings[self.split_index:])

        print("Time for full pipeline:  {:.2f}s".format(self.full_time))
        print("Time for slow pipeline:  {:.2f}s".format(self.slow_time))
        print("Time for fast pipeline:  {:.2f}s".format(self.fast_time))
        time_save_percent = 100-100*self.fast_time/self.full_time
        print("Time saving: {:.2f}%".format(time_save_percent))

        self.worth_splitting = time_save_percent > 10.

        if (not self.worth_splitting) and (self.first_fast_module is not None):
            print("")
            print("No significant time saving (<10%) from a fast-slow split.")
            print("But you told me specifically which module to use, so I will")
            print("split the pipeline anyway, though it may be slower.")
            print("You may wish to reconsider your choices, but I do what I'm told.")
            print("")
            self.worth_splitting = True
        elif not self.worth_splitting:
            print("")
            print("No significant time saving (<10%) from a fast-slow split.")
            print("Not splitting pipeline into fast and slow parts.")
            print("")
            self.split_index = len(timings)

        self.slow_modules = self.split_index
        self.fast_modules = len(pipeline.modules) - self.slow_modules
        self.slow_params = sum(list(first_use.values())[:self.split_index], [])
        self.fast_params = sum(list(first_use.values())[self.split_index:], [])

        if self.worth_splitting:
            print("")
            print("Based on this we have decided: ")
            print("   Slow modules ({}):".format(self.slow_modules))
            for module in pipeline.modules[:self.split_index]:
                print("        %s"% module.name)
            print("   Fast modules ({}):".format(self.fast_modules))
            for module in pipeline.modules[self.split_index:]:
                print("        {}".format(module.name))
            print("   Slow parameters ({}):".format(len(self.slow_params)))
            for param in self.slow_params:
                print("        {}--{}".format(*param))
            print("   Fast parameters ({}):".format(len(self.fast_params)))
            for param in self.fast_params:
                print("        {}--{}".format(*param))
            print("")
            print("")
        self.analyzed = True

    def _choose_fast_slow_split(self, first_use_count, timings, grid_mode):
        #some kind of algorithm to loop through the modules
        #and work out the time saving if we did the fast/slow from there.
        #at the moment just return the last module where there are any 
        #parameters
        #n = len(slow_count)
        if self.first_fast_module is not None:
            print("You manually told me in the parameter file to use module {} as the first in the fast block".format(self.first_fast_module))
            return self.first_fast_module


        n_step = len(timings)
        n_total = sum(first_use_count)
        T_total = sum(timings)

        if grid_mode:
            print("Analyzing the fast/slow for grid sampler - different rule used for time saving")

        T = np.zeros(n_step)
        for i in range(n_step):
            T_slow = sum(timings[:i])
            n_slow = sum(first_use_count[:i])
            T_fast = T_total - T_slow
            n_fast = n_total - n_slow
            if n_fast==0:
                T[i] = T_total
            elif grid_mode:
                T[i] = (T_slow/T_total)*10**(-n_fast) + (T_fast/T_total)*(1-10**(-n_fast))
            else:
                T[i] = (T_fast*n_fast)/(T_total*n_total) + (T_slow*n_slow)/(n_total*T_total)

        index = T.argmin()
        Tmin = T.min()
        index = (np.where(T==Tmin)[0]).max()
        print("Will use module {} as the first in the fast block".format(index))

        return index


class Pipeline(object):

    u"""A container of :class:`DataBlock`-processing :class:`Module`ʼs.

    Cosmosis consists of a large number of computational modules, which
    the user can arrange into pipelines in the configuration files
    before submitting a run.  This class encapsulates the basic concept
    of a pipeline, and provides methods for setting one up based on
    options and/or initial parameter vectors, and for running the
    pipeline to obtain a refinement of a :class:`DataBlock` (the
    prototypical use-case is Bayesian updating of prior likelihoods,
    though this is handled by the more specialized
    :class:`LikelihoodPipeline` defined below).

    Sometimes it will be possible to optimize execution of a pipeline by
    providing pre-emptive data which can be taken as the result of running
    the first few modules.  In this case a run will ‘bypass’ those modules
    and start the run at a `shortcut_module`, using the `shortcut_data` as
    the initial data set.

    """

    def __init__(self, arg=None, load=True, modules=None):

        u"""Pipeline constructor.

        The (poorly named) `arg` needs to be some reference to a parameter 
        :class:`Inifile` which includes all the information to form this
        pipeline: it can be a list of filenames (.ini files and such) to
        read for the parameters, or a :class:`Inifile` object directly.

        If `load` is `True` then all the modules in the pipelineʼs
        configuration will be loaded into memory and initialized.

        """
        if arg is None:
            arg = list()

        if isinstance(arg, config.Inifile):
            self.options = arg
        else:
            self.options = config.Inifile(arg)

        #This will be set later
        self.root_directory = self.options.get("runtime", "root", fallback=os.getcwd())

        self.quiet = self.options.getboolean(PIPELINE_INI_SECTION, "quiet", fallback=True)
        self.debug = self.options.getboolean(PIPELINE_INI_SECTION, "debug", fallback=False)
        self.timing = self.options.getboolean(PIPELINE_INI_SECTION, "timing", fallback=False)
        shortcut = self.options.get(PIPELINE_INI_SECTION, "shortcut", fallback="")
        if shortcut=="":
            shortcut=None

        self.do_fast_slow = self.options.getboolean(PIPELINE_INI_SECTION, "fast_slow", fallback=False)
        if self.do_fast_slow and shortcut:
            sys.stderr.write("Warning: you have the fast_slow and shortcut options both set, and we can only do one of those at once (we will do shortcut)\n")
            self.do_fast_slow = False
        self.slow_subspace_cache = None #until set in method
        self.first_fast_module = self.options.get(PIPELINE_INI_SECTION, "first_fast_module", fallback="")

        # initialize modules
        self.modules = []
        self.has_run = False
        if modules is not None:
            self.modules = modules

        elif load and PIPELINE_INI_SECTION in self.options.sections():
            module_list = self.options.get(PIPELINE_INI_SECTION,
                                           "modules", fallback="").split()
            self.modules = [
                module.Module.from_options(module_name,self.options,self.root_directory)
                for module_name in module_list
            ]
        else:
            self.modules = []

        if self.modules:
            self.shortcut_module=0
            self.shortcut_data=None
            if shortcut is not None:
                try:
                    index = module_list.index(shortcut)
                except ValueError:
                    raise ValueError("You tried to set a shortcut in "
                        "the pipeline but I do not know module %s"%shortcut)
                if index == 0:
                    print("You set a shortcut in the pipeline but it was the first module.")
                    print("It will make no difference.")
                else:
                    print("Shortcut mode activated: the first pipeline run will proceed as normal.")
                    print("Subsequent runs will use module {}, {}, as their first module".format(index,shortcut))
                    print("and use the cached results from the first run for everything before that.")
                    print("except the input parameter values. Think about this to check it's what you want.")
                self.shortcut_module = index
            self.shortcut_module=0
            self.shortcut_data=None



    def find_module_file(self, path):
        u"""Find a module file, which is assumed to be either absolute or relative to COSMOSIS_SRC_DIR"""
        return os.path.join(self.root_directory, path)



    def setup(self):
        u"""Run all of our modulesʼ `setup()` routines."""
        
        if self.timing:
            timings = [time.time()]

        for module in self.modules:
            # identify parameters needed for module setup
            relevant_sections = self.options.sections()

            #We let the user specify additional global sections that are
            #visible to all modules
            global_sections = self.options.get("runtime", "global", fallback=" ")
            for global_section in global_sections.split():
                relevant_sections.append(global_section)

            config_block = config_to_block(relevant_sections, self.options)
            config_block[PIPELINE_INI_SECTION, 'current_module'] = module.name
            # we need to store an ID here to allow a hack to add new parameters
            # from the modules themselves.  Unfortunately cosmosis stores ints, whereas
            # python IDs are long.  So we split the ID into two parts here

            LikelihoodPipeline.pipeline_being_set_up.append(self)
            LikelihoodPipeline.module_being_set_up.append(module.name)
            try:
                module.setup(config_block, quiet=self.quiet)
            finally:
                # This should only go wrong if someone is doing something
                # ridiculous, so I won't catch any error in it.
                LikelihoodPipeline.pipeline_being_set_up.remove(self)
                LikelihoodPipeline.module_being_set_up.remove(module.name)

            if self.timing:
                timings.append(time.time())

        if not self.quiet:
            sys.stdout.write("Setup all pipeline modules\n")

        if self.timing:
            timings.append(time.time())
            sys.stdout.write("Module timing:\n")
            for name, t2, t1 in zip(self.modules, timings[1:], timings[:-1]):
                sys.stdout.write("%s %f\n" % (name, t2-t1))


    def setup_fast_subspaces(self, all_params=False, grid=False):
        if self.do_fast_slow:
            print("Doing fast/slow parameter splitting")
            if self.first_fast_module:
                for i,module in enumerate(self.modules):
                    if module.name==self.first_fast_module:
                        first_fast_index=i
                        break
                else:
                    raise ValueError("You set first_fast_module={} but never ran that module".format(self.first_fast_module))
            else:
                first_fast_index = None

            self.slow_subspace_cache = SlowSubspaceCache(first_fast_module=first_fast_index)
            self.slow_subspace_cache.analyze_pipeline(self, all_params=all_params, grid=grid)

            if not self.slow_subspace_cache.worth_splitting:
                self.slow_subspace_cache = None
                self.do_fast_slow = False
                return


            self.fast_time = self.slow_subspace_cache.fast_time
            self.slow_time = self.slow_subspace_cache.slow_time
            # This looks a bit weird but makes sure that self.fast_params
            # and self.slow_params contain objects of type Parameter
            # not just the (section,name) tuples.
            if all_params:
                params = self.parameters
            else:
                params = self.varied_params
            self.fast_params = [p for p in params 
                if p in self.slow_subspace_cache.fast_params]
            self.slow_params = [p for p in params 
                if p in self.slow_subspace_cache.slow_params]
            self.fast_param_indices = [params.index(p)
                for p in self.fast_params]
            self.slow_param_indices = [params.index(p)
                for p in self.slow_params]
            self.n_fast_params = len(self.fast_params)
            self.n_slow_params = len(self.slow_params)
        else:
            self.slow_subspace_cache = None


    def cleanup(self):
        u"""Call every `module`ʼs `cleanup` method."""
        for module in self.modules:
            module.cleanup()



    def make_graph(self, data, filename):
        u"""Put a description of a graphical model in the graphviz format
        of a completed datablock `data' that was run on the pipeline, 
        in `filename`, illustrating the data flow of this pipeline.

        Graphviz tools can then be used to generate an actual image.

        The :class:`DataBlock` `data`ʼs attached log is used to determine
        the state of each module, and colourization is used in the graphic
        to indicate this.

        """
        try:
            import pygraphviz as pgv
        except ImportError:
            print("Cannot generate a graphical pipeline; please install the python package pygraphviz (e.g. with pip install pygraphviz)")
            return
        P = pgv.AGraph(directed=True)
        # P = pydot.Cluster(label="Pipeline", color='black',  style='dashed')
        # G.add_subgraph(P)
        def norm_name(name):
            return name #.replace("_", " ").title()
        known_sections = set()
        P.add_node("Sampler", color='Pink', style='filled', group='pipeline',shape='octagon', fontname='Courier')
        for module in self.modules:
            # module_node = pydot.Node(module.name, color='Yellow', style='filled')
            P.add_node(norm_name(module.name), color='lightskyblue', style='filled', group='pipeline', shape='box')
            known_sections.add(norm_name(module.name))
        P.add_edge("Sampler", norm_name(self.modules[0].name), color='lightskyblue', style='bold', arrowhead='none')
        for i in range(len(self.modules)-1):
            P.add_edge(norm_name(self.modules[i].name),norm_name(self.modules[i+1].name), color='lightskyblue', style='bold', arrowhead='none')
        # D = pydot.Cluster(label="Data", color='red', style='dashed')
        # G.add_subgraph(D)
        # #find
        log = [data.get_log_entry(i) for i in range(data.get_log_count())]
        for entry in log:
            if entry!="MODULE-START":
                section = entry[1]
                if section not in known_sections:
                    if section=="Results":
                        pass
                    else:
                        P.add_node(norm_name(section), color='yellow', style='filled', fontname='Courier', shape='box')
                    known_sections.add(section)
        module="Sampler"
        known_edges = set()
        for entry in log:
            if entry[0]=="MODULE-START":
                module=entry[1]
            elif entry[0]=="WRITE-OK" or entry[0]=="REPLACE-OK":
                section=entry[1]
                if (module,section,'write') not in known_edges:
                    P.add_edge(norm_name(module), norm_name(section), color='green')
                    known_edges.add((module,section,'write'))
            elif entry[0]=="READ-OK":
                section=entry[1]
                if (section,module,'read') not in known_edges:
                    P.add_edge((norm_name(section),norm_name(module)), color='grey50')
                    known_edges.add((section,module,'read'))

        P.write(filename)



    def run(self, data_package):
        u"""Run every module, in sequence, on DataBlock `data_package`.

        Apart from that the function goes to a lot of effort to provide
        run-time diagnostic information to the user.

        However, if any module returns anything but a zero status, the
        whole pipeline will cease to run and a :class:`ValueError` will be
        raised.

        Bear in mind the note on short-cuts in the class description
        above: if `shortcut_data` and `shortcut_module` are defined, then
        the pipeline will start at `shortcut_module` with `shortcut_data`
        as the initial parameter vector.

        """
        modules = self.modules
        if self.timing:
            self.timings = None

        timings = []
        if self.shortcut_module:
            if self.shortcut_data is None:
                first_module = 0
            else:
                first_module = self.shortcut_module
                existing_keys = set(data_package.keys())
                for (sec, name) in self.shortcut_data.keys():
                    if (sec,name) not in existing_keys:
                        data_package[sec, name] = self.shortcut_data[sec,name]
        elif self.slow_subspace_cache:
            first_module = self.slow_subspace_cache.start_pipeline(data_package)
            if first_module != 0 and (self.debug or self.timing):
                sys.stdout.write("COOL: Quickstarting pipeline from module {} (fast/slow)\n".format(first_module))
                sys.stdout.flush()
        else:
            first_module = 0

        if self.timing:
            start_time = time.time()

        for module_number, module in enumerate(modules):
            if module_number<first_module:
                continue
            if self.debug:
                sys.stdout.write("Running %.20s ...\n" % module)
                sys.stdout.flush()
            data_package.log_access("MODULE-START", module.name, "")
            if self.timing:
                t1 = time.time()

            status = module.execute(data_package)

            if status is None:
                raise ValueError(("A module you ran, '{}', did not return a proper status value.\n"+
                    "It should return an integer, 0 if everything worked.\n"+
                    "Sorry to be picky but this kind of thing is important.").format(module))

            if self.debug:
                sys.stdout.write("Done %.20s status = %d \n" % (module,status))
                sys.stdout.flush()

            if self.timing:
                t2 = time.time()
                timings.append(t2-t1)
                sys.stdout.write("%s took: %.3f seconds\n"% (module,t2-t1))

            if status:
                if self.debug:
                    data_package.print_log()
                    sys.stdout.flush()
                    sys.stderr.write("Because you set debug=True I printed a log of "
                                     "all access to data printed above.\n"
                                     "Look for the word 'FAIL' \n")
                    sys.stderr.write("Though the error message could also be somewhere above that.\n\n")
                if not self.quiet:
                    sys.stderr.write("Error running pipeline (%d)- "
                                     "hopefully printed above here.\n"%status)
                    sys.stderr.write("Aborting this run and returning "
                                     "error status.\n")
                    if not self.debug:
                        sys.stderr.write("Setting debug=T in [pipeline] might help.\n")
                return None

            # If we are using a fast/slow split (and we are not already running on a cached subset)
            # Then see if it wants to cache these results
            if self.slow_subspace_cache and first_module==0:
                self.slow_subspace_cache.next_module_results(module_number, data_package)

            # Alternatively we will do the shortcut thing
            elif self.shortcut_module and (not self.has_run) and module_number==self.shortcut_module-1:
                print("Saving shortcut data")
                self.shortcut_data = data_package.clone()

        if self.timing:
            end_time = time.time()
            sys.stdout.write("Total pipeline time: {:.3} seconds\n".format(end_time-start_time))
            self.timings = timings

        if not self.quiet:
            sys.stdout.write("Pipeline ran okay.\n")

        data_package.log_access("MODULE-START", "Results", "")
        # return something
        self.has_run = True
        return True

    def clear_cache(self):
        self.slow_subspace_cache.clear_cache()




class LikelihoodPipeline(Pipeline):

    u"""Very specialized pipeline designed specifically for the prototypical case of Bayes-computed posterior distributions.

    The point of a statistical updating pipeline is that the parameters in
    the datablocks passed down the pipe, as well as having currently
    estimated values, also have allowable ranges and possibly other
    constraints which the user may want to tinker before each run.  Thus
    there is a specialized layout of initialization files in the file
    system, and there is a modified expectation on the modules to perform
    simulation, compute the Bayesian evidence, hence log-likelihood.  The
    pipeline itself will aggregate the results and summarize the net
    effect of all the likelihood estimations, and thence compute the
    Bayesian posterior.

    Because of the necessity of working with distributions of values for
    each parameter, rather than just a scalar, the extra information is
    stored in a shadow array—another dictionary with the same keys but a
    complementary set of values to the original ones—of
    :class:`parameter`s to the :class:`datablock` which the base pipeline
    modifies (actually only a subset of them known as the `varied_params`:
    an array which references the interesting parameters in the full set).
    this shadow array (`parameters`) is often referred to simply as `p`,
    and the two arrays frequently need to be ‘zipped’ together and then
    ‘unzipped’ after computations have completed.

    """
    # These class-level variables are used by the
    # register_new_parameter function.
    pipeline_being_set_up = []
    module_being_set_up = []

    def __init__(self, arg=None, id="", override=None, modules=None, load=True, values=None, priors=None, only=None):
        u"""Construct a :class:`LikelihoodPipeline`.

        The arguments `arg` and `load` are used in the base-class
        initialization (see above).  The `id` is given to our `id_code`
        (which doesnʼt seem to have a purpose), and `override` is a
        dictionary of `(section, name)->value` which will override any
        settings for those parametersʼ values in the initialization files.
        
        """
        super().__init__(arg=arg, load=load, modules=modules)

        if id:
            self.id_code = "[%s] " % str(id)
        else:
            self.id_code = ""
        self.n_iterations = 0


        if values is None:
            self.values_file = self.options.get(PIPELINE_INI_SECTION, "values", fallback="")
            self.values_filename = self.values_file
        else:
            self.values_file = values
            self.values_filename = None

        if priors is None:
            priors_files = self.options.get(PIPELINE_INI_SECTION,
                                            "priors", fallback="").split()
            self.priors_files = priors_files
        else:
            if isinstance(priors, config.Inifile):
                priors = [priors]
            self.priors_files = priors
        self.parameters = parameter.Parameter.load_parameters(self.values_file,
                                                              self.priors_files,
                                                              override,
                                                              )
        # We set up the modules first, so that if they want to e.g.
        # add parameters then they can.
        self.setup()

        if only:
            for p in only:
                if p not in self.parameters:
                    raise ValueError("You used --only but the parameter "
                        "you named ({}) is not in the pipeline".format(p))
            for p in self.parameters:
                if p not in only:
                    p.fix()

        self.reset_fixed_varied_parameters()

        self.print_priors()

        #We want to save some parameter results from the run for further output
        extra_saves = self.options.get(PIPELINE_INI_SECTION,
                                       "extra_output", fallback="")

        self.extra_saves = []
        self.number_extra = 0
        for extra_save in extra_saves.split():
            section, name = extra_save.upper().split('/')
            self.extra_saves.append((section, name))
            # To load an array-type extra_output, we should know beforehand its size
            # So, it reads from the name especification
            # e.g. data_vector/2pt_theory#457, in which case #457 specifies the size
            self.number_extra += int(name.split('#')[-1]) if '#' in name else 1


        #pull out all the section names and likelihood names for later

        likelihood_names = self.options.get(PIPELINE_INI_SECTION,
                                            "likelihoods",
                                            fallback=NO_LIKELIHOOD_NAMES)
        if likelihood_names==NO_LIKELIHOOD_NAMES:
            self.likelihood_names = NO_LIKELIHOOD_NAMES
        else:
            self.likelihood_names = likelihood_names.split()




    def print_priors(self):
        u"""Pretty-print a table of priors for human inspection."""
        print("")
        print("Parameter Priors")
        print("----------------")
        if self.parameters:
            n = max([len(p.section)+len(p.name)+2 for p in self.parameters])
        else:
            n=1
        for param in self.parameters:
            s = "{}--{}".format(param.section,param.name)
            print("{0:{1}}  ~ {2}" .format(s, n, param.prior))
        print("")

    def _register_new_parameter(self,
                                module_name,
                                section,
                                name,
                                start,
                                min_value,
                                max_value,
                                prior_name,
                                prior_args):
        # This is designed to be called by the register_new_parameter
        # function below, from modules themselves, to support the case
        # where modules create their own parameter
        if max_value == min_value:
            limits = None
        else:
            limits = (min_value, max_value)
        if prior_name == "":
            prior_obj = None
        else:
            if prior_args is None:
                prior_args = []
            prior_text = prior_name + " " + " ".join([str(x) for x in prior_args])
            prior_obj = prior.Prior.parse_prior(prior_text)
        param = parameter.Parameter(section, name, start, limits, prior_obj)
        print("Pipeline module {} created new parameter {}".format(module_name, param))
        print("    with start:", param.start)
        print("    with limits:", param.limits)
        print("    with prior:", param.prior)
        self.parameters.append(param)

    def reset_fixed_varied_parameters(self):
        u"""Identify the sub-set of parameters which are fixed, and those which are to be varied."""
        self.varied_params = [param for param in self.parameters
                              if param.is_varied()]
        self.fixed_params = [param for param in self.parameters
                             if param.is_fixed()]
        self.nvaried = len(self.varied_params)
        self.nfixed = len(self.fixed_params)



    def parameter_index(self, section, name):
        u"""Return the sequence number of the parameter `name` in `section`.

        If the parameter is not found then :class:`ValueError` will be raised.

        """
        i = self.parameters.index((section, name))
        if i==-1:
            raise ValueError("Could not find index of parameter %s in section %s"%(name, section))
        return i



    def set_varied(self, section, name, lower, upper):
        u"""Indicate that the parameter (`section`, `name`) is to be varied between the `lower` and `upper` bounds."""
        i = self.parameter_index(section, name)
        self.parameters[i].limits = (lower,upper)
        self.reset_fixed_varied_parameters()



    def set_fixed(self, section, name, value):
        u"""Indicate that the parameter (`section`, `name`) must be held fixed at `value`."""
        i = self.parameter_index(section, name)
        self.parameters[i].limits = (value, value)
        self.reset_fixed_varied_parameters()



    def output_names(self):
        u"""Return a list of strings, each the name of a non-fixed parameter."""
        param_names = [str(p) for p in self.varied_params]
        extra_names = []
        for section,name in self.extra_saves:
            if ('#' in name):
                n,l = name.split('#')
                for i in range(l):
                    extra_names.append('{}--{}_{}'.format(section,n,i))
            else:
                extra_names.append('%s--%s'%(section,name))
        return param_names + extra_names



    def randomized_start(self):
        u"""Give each varied parameter an independent random value distributed according
        to the parameter prior.

        The return is a `NumPy` :class:`array` of the random values.

        """
        
        # should have different randomization strategies
        # (uniform, gaussian) possibly depending on prior?
        
        return np.array([p.random_point() for p in self.varied_params])



    def is_out_of_range(self, p):
        u"""Determine if any parameter is not in its allowed range."""
        return any([not param.in_range(x) for
                    param, x in zip(self.varied_params, p)])



    def denormalize_vector(self, p, raise_exception=True):
        u"""Convert an array of normalized parameter values, one for each varied parameter,
        in the range [0.0,1.0] into their original values using only the lower and upper limits of the parameter.
    
        Use denormalize_vector_from_prior to convert according to the prior instead.

        """
        return np.array([param.denormalize(x, raise_exception) for param, x
                         in zip(self.varied_params, p)])



    def denormalize_vector_from_prior(self, p):
        r"""Convert an array of normalized parameter values, one for each varied parameter,
        in the range [0.0,1.0] into their original values according to the prior for each parameter.

        i.e. 
        v -> x  such that \int_{-inf}^{x} p(x') dx' = v

        """
        return np.array([param.denormalize_from_prior(x) for param, x
                         in zip(self.varied_params, p)])



    def normalize_vector(self, p):
        u"""Convert an array of parameter values, one for each varied parameter,
         into a normalized form all in the range [0.0,1.0] using only the lower and upper limits for each parameter."""
        return np.array([param.normalize(x) for param, x
                         in zip(self.varied_params, p)])



    def normalize_matrix(self, c):
        u"""Roughly, return a correlation matrix corresponding to the 
        covariance matrix `c`, of `varied_params` values.

        Except that the elements of `c` are not probabilities but
        dimensional values, and the ‘normalization’ is relative to the
        range of values the ‘covariance’s can take given the lower and
        upper limits on the variates.

        """
        c = c.copy()
        n = c.shape[0]
        assert n==c.shape[1], "Cannot normalize a non-square matrix"
        for i in range(n):
            pi = self.varied_params[i]
            ri = pi.limits[1] - pi.limits[0]
            for j in range(n):
                pj = self.varied_params[j]
                rj = pj.limits[1] - pj.limits[0]
                c[i,j] /= (ri*rj)
        return c



    def denormalize_matrix(self, c, inverse=False):
        u"""Perform the inverse operation to the normalize_matrix function above.

        Note that if `inverse` is `True` the action is *exactly* the same
        as the function above, i.e. it *normalizes* the matrix.

        """
        c = c.copy()
        n = c.shape[0]
        assert n==c.shape[1], "Cannot normalize a non-square matrix"
        for i in range(n):
            pi = self.varied_params[i]
            ri = pi.limits[1] - pi.limits[0]
            for j in range(n):
                pj = self.varied_params[j]
                rj = pj.limits[1] - pj.limits[0]
                if inverse:
                    c[i,j] /= (ri*rj)
                else:
                    c[i,j] *= (ri*rj)
        return c



    def start_vector(self, all_params=False, as_array=True):
        u"""Return a vector of starting values for parameters.

        If `all_params` is specified as `True` then the return will include all
        our `parameters`, otherwise only the varying ones are included.

        If `as_array` is specified as `False` then a Python list is
        returned, otherwise, the default, a NumPy array is returned.

        """
        if all_params:
            p = [param.start for param in self.parameters]
        else:            
            p =[param.start for param in self.varied_params]
        if as_array:
            p = np.array(p)
        return p



    def min_vector(self, all_params=False):
        u"""Return a NumPy array of lower limits for the parameters in the pipeline.

        If `all_params` is specified as `True` then the return will 
        include all parameters, including fixed ones. Otherwise it will just be the 
        varying parameters.

        """
        if all_params:
            return np.array([param.limits[0] for
                 param in self.parameters])
        else:
            return np.array([param.limits[0] for
                         param in self.varied_params])



    def max_vector(self, all_params=False):
        u"""Return a NumPy array of upper limits for the parameters in the pipeline.

        If `all_params` is specified as `True` then the return will 
        include all parameters, including fixed ones. Otherwise it will just be the 
        varying parameters.

        """
        if all_params:
            return np.array([param.limits[1] for
                 param in self.parameters])
        else:
            return np.array([param.limits[1] for
                         param in self.varied_params])

    def build_starting_block(self, p, check_ranges=False, all_params=False):
        u"""Assemble :class:`DataBlock` data based on parameter values in `p`, and return it.


        If `check_ranges` is indicated, the function will return `None` if
        **any** of our parameters are out of their indicated range.

        If `all_params` is indicated, then the `p` run data will be
        assumed to match all the pipeline parameter, including fixed ones.
        Otherwise (the default) it should match the list ‘varied_params’, and all
        of our ‘fixed’ parameters are added to the run-set.

        """
        if check_ranges:
            if self.is_out_of_range(p):
                return None

        data = block.DataBlock()

        if all_params:
            for param, x in zip(self.parameters, p):
                data[param.section, param.name] = x
        else:
            # add varied parameters
            for param, x in zip(self.varied_params, p):
                data[param.section, param.name] = x

            # add fixed parameters
            for param in self.fixed_params:
                data[param.section, param.name] = param.start

        return data


    def run_parameters(self, p, check_ranges=False, all_params=False):
        u"""Assemble :class:`DataBlock` data based on parameter values in `p`, and run the pipeline on those data.

        If `check_ranges` is indicated, the function will return `None` if
        **any** of our parameters are out of their indicated range.

        If `all_params` is indicated, then the `p` run data will be
        assumed to match all the pipeline parameter, including fixed ones.
        Otherwise (the default) it should match the list ‘varied_params’, and all
        of our ‘fixed’ parameters are added to the run-set.

        """
        data = self.build_starting_block(p, check_ranges=check_ranges, all_params=all_params)

        if self.run(data):
            # First run.  If we have not set the likelihood names in the parameter
            # file then get them from the block
            if self.likelihood_names == NO_LIKELIHOOD_NAMES:
                self._set_likelihood_names_from_block(data)
                
            return data
        else:
            sys.stderr.write("Pipeline failed on these parameters: {}\n".format(p))
            return None



    def create_ini(self, p, filename):
        u"""Dump the parameters `p` as a new ini file at `filename`"""
        output = collections.defaultdict(list)
        for param, x in zip(self.varied_params, p):
            output[param.section].append("%s  =  %r    %r    %r\n" % (
                param.name, param.limits[0], x, param.limits[1]))
        for param in self.fixed_params:
            output[param.section].append("%s  =  %r\n" % (param.name, param.start))
        ini = open(filename, 'w')
        for section, params in sorted(output.items()):
            ini.write("[%s]\n"%section)
            for line in params:
                ini.write(line)
            ini.write("\n")
        ini.close()



    def prior(self, p, all_params=False, total_only=True):
        u"""Compute the probability of all values in `p` based on their prior distributions.

        The array `p` should match the length of of all of our `parameters` if
        `all_params` is `True`, and our `varied_params` otherwise.

        If `total_only` is `True` (the default), then the scalar sum of
        all the prior probabilities is returned.  Otherwise a list
        of pairs is returned, with each element a stringified version of
        the parameter name, and the prior probability:
        [(name1, prior1), (name2,prior2), ...]

        """
        if all_params:
            params = self.parameters
        else:
            params = self.varied_params
        priors = [(str(param),param.evaluate_prior(x)) for param,x in zip(params,p)]
        if total_only:
            return sum(pr[1] for pr in priors)
        else:
            return priors

    def run_results(self, p, all_params=False):
        u"""Run the pipeline on the given parameters and get a results object.

        The argument `p` is a vector or list of the input parameters.
        if (as in the default) `all_params` is False, then it should be the
        same length and order as `self.varied_params`.

        Otherwise, if all_params is False, then it should match the length and order
        of `self.parameters`.

        The method returns a `PipelineResults` object with the following attributes:

        * results.post, the posterior (float);

        * results.extra, the reqiured additional output parameters (numpy array)

        * results.prior, the total prior (float)

        * results.block, the updated data block

        If there is a problem anywhere in the computations which does
        *not* cause a run-time exception to be raised—including the case
        where a parameter goes outside of its alloted range—, then
        `-numpy.inf` will be returned as the final posterior (i.e., zero
        probability of this set of parameter values being correct).

        """
        r = PipelineResults(p, self.number_extra)

        priors = self.prior(p, all_params=all_params, total_only=False)
        r.prior = sum(pr[1] for pr in priors)

        if np.isnan(r.prior):
            r.prior = -np.inf

        if not np.isfinite(r.prior):
            if not self.quiet:
                print("Proposed outside bounds: prior -infinity")
            return r
        try:
            like, r.extra, r.block = self.likelihood(p, return_data=True, all_params=all_params)
            r.set_like(like)

            if r.block is not None:
                for name,pr in priors:
                    r.block["priors", name] = pr

        except Exception:
            if self.debug:
                sys.stderr.write("\n\nERROR: there was an exception running the likelihood:\n")
                sys.stderr.write("\nBecause you have debug=T I will let this kill the chain.\n")
                sys.stderr.write("The input parameters were:{}\n".format(repr(p)))
                raise

            sys.stderr.write("\n\nERROR: there was an exception running the likelihood:\n")
            sys.stderr.write("The input parameters were:{}\n".format(repr(p)))
            traceback.print_exc(file=sys.stderr)
            sys.stderr.write("You should fix this but for now I will return NaN for the likelihood (because you have debug=F)\n\n")

        if np.isnan(r.post):
            r.post = -np.inf

        if np.isnan(r.like):
            r.like = -np.inf

        return r




    def posterior(self, p, return_data=False, all_params=False):
        u"""Use the above methods to obtain prior and updated log-likelihoods, sum together to get Bayesian posterior.

        The argument `p` is a vector or list of the input parameters.
        if (as in the default) `all_params` is False, then it should be the
        same length and order as `self.varied_params`.

        Otherwise, if all_params is False, then it should match the length and order
        of `self.parameters`.
        The method returns two or three values depending on `return_data`:

        * The posterior;

        * a vector (NumPy array) of updated parameter values as specified
          in `self.extra_saves`;

        * if `return_data` was specified, the updated data block.

        If there is a problem anywhere in the computations which does
        *not* cause a run-time exception to be raised—including the case
        where a parameter goes outside of its alloted range—, then
        `-numpy.inf` will be returned as the final posterior (i.e., zero
        probability of this set of parameter values being correct).

        """

        r = self.run_results(p, all_params=all_params)
        if return_data:
            return r.post, r.extra, r.block
        else:
            return r.post, r.extra



    def _set_likelihood_names_from_block(self, data):
        likelihood_names = []
        for _,key in data.keys(section_names.likelihoods):
            if key.endswith("_like"):
                name = key[:-5]
                likelihood_names.append(name)
        self.likelihood_names = likelihood_names

        if not self.quiet:
            # Tell the user what we found.
            print("Likelihoods not set in parameter file, so checking what is generated:")
            for name in self.likelihood_names:
                print("Found likelihood named {}".format(name))
            if not self.likelihood_names:
                print("No likelihoods found")

    def _extract_likelihoods(self, data):
        "Extract the likelihoods from the block"

        section_name = section_names.likelihoods

        # loop through named likelihoods and sum their values
        likelihoods = []
        for likelihood_name in self.likelihood_names:
            try:
                L = data.get_double(section_name,likelihood_name+"_like")
                likelihoods.append(L)
                if not self.quiet:
                    print("    Likelihood {} = {}".format(likelihood_name, L))
            # Complain if not found
            except block.BlockError:
                raise MissingLikelihoodError(likelihood_name, data)

        # Total likelihood
        like = sum(likelihoods)

        # DM: Issue #181: Zuntz: replace NaN's with -inf's in posteriors and
        #                 likelihoods.
        if np.isnan(like):
            like = -np.inf

        if not self.quiet and self.likelihood_names:
            sys.stdout.write("Likelihood total = {}\n".format(like))

        return like

        
    def likelihood(self, p, return_data=False, all_params=False):
        u"""Run the simulation pipeline, computing any log-likelihoods in the pipeline 
        given the given input parameter values, and return the sum of these.

        The parameter vector `p` must match the length of `self.varied_params`,
        unless `all_params` is specified as `True` in which case it must match
        `self.parameters', i.e. must correspond to the complete parameter set.

        If `return_data` are requested, then the updated data block will
        be returned as the third return item.

        The return will consist of two or three items, depending on
        `return_data`:

          * A scalar holding the sum of all computed log-likelihoods of
            the updated parameter value vector;

          * a vector (NumPy array) of updated parameter values as
            specified in `self.extra_saves`;

          * if `return_data` was specified, the updated data block.

        If anything goes wrong in any of the computation which does *not*
        result in a run-time error being raised (which would include the
        case of a parameter going outside of its stipulated limits), then
        the returned log-likelihood will be `-np.inf`.

        """

        data = self.run_parameters(p, all_params=all_params)
        if data is None:
            if return_data:
                return -np.inf, [np.nan for i in range(self.number_extra)], data
            else:
                return -np.inf, [np.nan for i in range(self.number_extra)]

        like = self._extract_likelihoods(data)

        extra_saves = []
        for option in self.extra_saves:
            try:
                #JAZ - should this be just .get(*option) ?
                # If the # symbol is present, we have an array-type extra_output and it is
                # read as numpy.ndarray. We append all of its elements to the extra_saves.
                # If the total array size does not match the sum of the sizes indicated in the
                # name #, an exception will be raised.
                if('#' in option[1]):
                    value = data.get(option[0], option[1].split('#')[0])
                else:
                    value = data.get(*option)

            except block.BlockError:
                value = np.nan

            if (type(value)==np.ndarray):
                for e in value:
                    extra_saves.append(e)
            else:
                extra_saves.append(value)
                # ---------------------------- otavio end ---------------------------

        self.n_iterations += 1
        if return_data:
            return like, extra_saves, data
        else:
            return like, extra_saves



def config_to_block(relevant_sections, options):
    u"""Compose :class:`DataBlock` of parameters only in `relevant_sections` of complete set of `options`."""
    config_block = block.DataBlock()

    for (section, name), value in options:
        if section in relevant_sections:
            # add back a default section?
            val = options.gettyped(section, name)
            if val is not None:
                config_block.put(section, name, val)
    # For book-keeping stuff later we also need to record
    # which keys were in the default section.  If
    # we want to keep track of unused parameters in
    # Module.access_check_report then we need to ensure
    # we don't include parameters that are only in the section
    # because they are in every section.  So we keep track of
    # such parameters by including them here
    for (key, val) in options.items("DEFAULT"):
        config_block.put("_cosmosis_default_section", key, val)
    return config_block



#coding: utf-8


u"""Definition of the :class:`Parameter` class."""
import random
from . import config
from . import prior as priors # to avoid breaking other stuff below
import numpy as np



class Parameter(object):

    u"""Distribution meta-data for a :class:`Pipeline` parameter.
    
    While pipeline modules deal with a dictionary of scalar parameters,
    the pipeline infrastructure (i.e., the :class:`Pipeline` class) must
    track meta-data for each parameter: the allowable range of values the
    parameter is allowed to take, and the (prior) distribution from which
    the values should be sampled.  These data, *but not the parameter
    values themselves*, are stored in this class and kept by the
    :class:`Pipeline` class in an array which parallels the
    :class:`DataBlock` or :class:`Config` which holds the values.

    """

    def __init__(self, section, name, start, limits=None, prior=None):
        u"""Store meta-data for parameter at `(section, name)`.

        If `prior` is not supplied, a uniform distribution is assumed.

        If `limits` are not supplied, the parameter will be considered
        fixed at the `start` value.

        For the available priors, see the :module:Priors module.

        """
        if not limits:
            self.limits = (start, start)
        else:
            self.limits = limits
            if (limits[1]<limits[0]):
                raise ValueError("Your ini file specified that "
                    "parameter %s in section %s had upper limit "
                    "< lower limit"% (name, section))

        self.section = section
        self.name = name
        self.start = start

        if prior is None:
            if limits is None:
                # Parameter has no setting in the priors file and is fixed
                prior = priors.DeltaFunctionPrior(start)
            else:
                # Parameter has no setting in the priors file and is
                # variable
                prior = priors.UniformPrior(limits[0], limits[1])
        else:
            if limits is None:
                # Parameter does have setting in the priors file but is
                # fixed - just fix value
                prior = priors.DeltaFunctionPrior(start)
            else:
                # Parameter does have setting in the priors file and is
                # variable - truncate prior to limits
                prior = prior.truncate(limits[0], limits[1])

        self.prior = prior

        # TODO: check consistency of prior with limits



    def __eq__(self, other):
        u"""Return `True` if `other` stands for the same data block entry as us.

        The same entry means that the `(section, name)` pairs are the
        same.  Note that this is NOT a test of equality of the
        `Parameter`sʼ values!

        """
        if isinstance(other, (list, tuple)):
            try:
                section, name = other
            except ValueError:
                return False
            return (self.section == section and
                    self.name == name)
        elif isinstance(other, Parameter):
            return (self.section == other.section and
                    self.name == other.name)
        elif isinstance(other, str):
            return other==self.__str__()
        else:
            raise NotImplementedError("Tried to do parameter==something where something was not a thing I understand.")



    def __str__(self):
        u"""Return our ID "section--name" as stringified version."""
        return self.section + "--" + self.name



    def __repr__(self):
        u"""Return __str__."""
        return self.__str__()

    def fix(self, value=None):
        """
        Switch this parameter to being fixed, optionally to
        a new specified value.
        """
        if value is not None:
            self.start = value

        self.limits = self.start, self.start

    def is_fixed(self):
        u"""Test whether this parameter is fixed or varied.

        Returns `True` if parameter is fixed to a single value by
        degenerate definition of the limits, and false if there is room
        for it to vary.

        """
        return self.limits[0] == self.limits[1]



    def is_varied(self):
        u"""The opposite of :func:`is_fixed`."""
        return not self.is_fixed()



    def in_range(self, p):
        u"""Check that the value `p` is in this parameterʼs allowed range."""
        return self.limits[0] <= p <= self.limits[1]



    def width(self):
        u"""Return the difference between the upper and lower limits."""
        return self.limits[1] - self.limits[0]



    def random_point(self):
        u"""Return a random number taken from the 'prior' distribution."""
        return self.prior.sample(1)[0]



    def normalize(self, p):
        u"""Return the relative position of `p` between the allowable limits: a value fron 0.0 to 1.0.

        But the return value will go beyond the unit interval if `p` is
        actually outside the limits.

        """
        if self.is_fixed():
            return 0.0
        else:
            return (p - self.limits[0]) / (self.limits[1] - self.limits[0])



    def denormalize(self, p, raise_exception=True):
        u"""Return the value at the relative position `p` between the lower and upper limits.

        If `p` is outside the range [0.0, 1.0], then if `raise_exception`
        is `True` a ValueError will be raised, otherwise a value extrapolated
        outside the range of the limits will be returned.

        """
        if 0.0 <= p <= 1.0:
            return p*(self.limits[1]-self.limits[0]) + self.limits[0]
        elif not raise_exception:
            return p*(self.limits[1]-self.limits[0]) + self.limits[0]
        else:
            raise ValueError("parameter value {} for {} not normalized".format(p,self))



    def denormalize_from_prior(self, p):
        u"""Take `p` as a probability, and find the value which has that (cumulated) probability in the prior distribution."""
        if 0.0 <= p <= 1.0:
            return self.prior.denormalize_from_prior(p)
        else:
            raise ValueError("parameter value {} for {} not normalized".format(p,self))



    def evaluate_prior(self, p):
        u"""Get the probability of `p` coming from the prior distribution."""
        if p < self.limits[0] or p > self.limits[1]:
            return -np.inf
        elif self.prior:
            return self.prior(p)
        else:
            return 0.0



    @staticmethod
    def load_parameters(value_file, priors_files=None, override=None):
        u"""Return array of :class:`Parameters` as directed by the input files.

        Every key in the `value_file` will produce an entry in the
        returned array, with the inferred starting value, and lower and
        upper limits (space-separated list).

        Where a key is also found in a given file in `priors_files` then a
        :class:`Prior` object will be constructed and attached to the
        parameter.

        If `override` contains a corresponding `(section, name)` key, then
        it will provide the start and limit values for the parameter,
        regardless of the file contents.

        """
        if isinstance(value_file, config.Inifile):
            values_ini = value_file
        else:
            values_ini = config.Inifile(value_file)

        if priors_files:
            priors_data = priors.Prior.load_priors(priors_files)
        else:
            priors_data = {}

        parameters = []
        for (section, name), value in values_ini:
            #override if available
            if (override is not None) and (section, name) in override:
                value = override[(section,name)]

            # parse value line
            start, limits = Parameter.parse_parameter(value)

            # check for prior
            pri = priors_data.get((section, name), None)


            parameters.append(Parameter(section, name,
                                        start, limits, pri))

        return parameters



    @staticmethod
    def parse_parameter(line):
        u"""Interpret a `line` of one to three numbers as the start and range of a parameter.

        In all cases, the return will be a scalar start value and 2-tuple
        of limits.

        With one number, the start value will be taken and `None` returned
        as the limits tuple, indicating that the parameter is to be kept
        constant, equivalently a delta-function distribution.

        With two numbers, they will be taken as the limits and the
        starting value will be half way between the two.

        With three numbers, they will be taken directly as the start
        value, lower limit, and upper limit.

        """
        try:
            values = [float(p) for p in line.split()]
            if len(values) == 1:
                if values[0]==int(values[0]):
                    try:
                        v = int(line)
                        return v, None
                    except ValueError:
                        return values[0], None
                return values[0], None
            elif len(values) == 2:
                return 0.5*(values[0]+values[1]), tuple(values)
            elif len(values) == 3:
                return values[1], (values[0], values[2])
            else:
                raise ValueError("Was expecting 1-3 values for "
                                 "parameter value %s" % (line,))
        except ValueError as error:
            raise ValueError("Unable to parse numeric value for "
                             "parameter value %s, error %s" %
                             (line, error))


def register_new_parameter(options,
                           section,
                           name,
                           min_value,
                           start_value,
                           max_value,
                           prior_name="",
                           prior_args=None):
    """
    Add a parameter to the pipeline from a module setup function.

    We allow modules to define and add parameters to the pipeline,
    either fixed or sampled.  This function can be called to do this.

    It will only affect the pipeline from which setup was called.

    options is the datablock passed to setup(options).
    section is the string setting the datablock section for the new parameter
    name is the name of the new parameter
    min_value is the lower limit of the parameter's range
    max_value is the upper limit of the parameter's range
    prior_name (optional) is a string - see prior.py
    prior_args (optional) is any additional arguments to create the prior
    """
    from .pipeline import LikelihoodPipeline

    n = len(LikelihoodPipeline.pipeline_being_set_up)

    if n == 0:
        print("Tried to register_new_parameter from a module not in a pipeline: skipping.")
        return
    elif n != 1:
        raise RuntimeError("Multiple pipelines are currently being set up, cannot register new parameter")

    if len(LikelihoodPipeline.module_being_set_up) != n:
        raise RuntimeError("Internal error: not clear what module is being set up.")

    # Now we have established the numbers are right we
    # register the parameter
    pipeline = LikelihoodPipeline.pipeline_being_set_up[0]
    caller = LikelihoodPipeline.module_being_set_up[0]


    pipeline._register_new_parameter(caller,
                           section,
                           name,
                           start_value,
                           min_value,
                           max_value,
                           prior_name,
                           prior_args)



from .. import ParallelSampler
from . import pmc
import numpy as np

#needs to be global for MPI to behave
#I think this is still true, at least
global pipeline

def posterior(p):
    # PMC code needs posterior not log posterior
    r = pipeline.run_results(p)
    return r.post, (r.prior, r.extra)


class PmcSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("component", int), ("prior", float), ("post",float), ("log_weight",float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline
        self.n_iterations = self.read_ini("iterations", int, default=30)
        self.n_components = self.read_ini("components", int, default=5)
        self.grace_period = self.read_ini("grace_period", int, default=3)
        self.n_samples = self.read_ini("samples_per_iteration", int, 
            default=1000)
        self.final_samples = self.read_ini("final_samples", int, 
            default=5000)

        #Student's t mode
        student = self.read_ini("student", bool, default=False)
        if student:
            print()
            print("WARNING")
            print()
            print("Student's t mode probably not working yet")
            print("Unless you are testing you should probably set")
            print("student=F")
            print()
            print()
            nu = self.read_ini("nu", float, default=2.0)
        else:
            nu = None


        #start values from prior
        start = self.pipeline.start_vector()
        covmat = self.load_covariance_matrix()

        #Sampler object itself.
        quiet = self.pipeline.quiet
        self.sampler = pmc.PopulationMonteCarlo(posterior, self.n_components, 
            start, covmat, quiet=quiet, student=student, nu=nu, pool=self.pool)

        self.interrupted = False
        self.iterations = 0
        self.samples = 0

    def execute(self):
        if self.iterations==self.n_iterations:
            n = self.final_samples
            update=False
        else:
            n = self.n_samples
            update=True
        do_kill = self.iterations>=self.grace_period
        #Run the MCMC  sampler.
        try:
            results = self.sampler.sample(n,update, do_kill)
            #returns samples, like, extra, weights
        except KeyboardInterrupt:
            self.interrupted=True
            return
            
        self.iterations += 1
        self.samples += n

        for (vector, post, extra, component, weight) in zip(*results):
            prior, extra = extra
            self.output.parameters(vector, extra, (component, prior, post, weight))

        print("Done %d iterations, %d samples" % (self.iterations, self.samples))


    def is_converged(self):
         # user has pressed Ctrl-C
        if self.interrupted:
            print("Interrupted...")
            return True
        if self.iterations >= self.n_iterations+1:
            print("Full number of samples generated; sampling complete")
            self.output.final("nsample", self.final_samples)

            return True
        return False



    def load_covariance_matrix(self):
        covmat_filename = self.read_ini("covmat", str, "").strip()

        #No covariance specified - just use 1% of the range of the parameter.
        # This is the std. dev. here and is squared below.
        if covmat_filename == "":
            covmat = np.array([p.width()/100.0 for p in self.pipeline.varied_params])
        elif not os.path.exists(covmat_filename):
            raise ValueError(
            "Covariance matrix %s not found" % covmat_filename)
        else:
            covmat = np.loadtxt(covmat_filename)

        if covmat.ndim == 0:
            covmat = covmat.reshape((1, 1))
        elif covmat.ndim == 1:
            covmat = np.diag(covmat ** 2)

        nparams = len(self.pipeline.varied_params)
        if covmat.shape != (nparams, nparams):
            raise ValueError("The covariance matrix was shape (%d x %d), "
                    "but there are %d varied parameters." %
                    (covmat.shape[0], covmat.shape[1], nparams))
        return covmat

# legacy alias
PMCSampler = PmcSampler



import sys

fortran_template = """
! This module is auto-generated from the file cosmosis/datablock/section_names.txt.
! Edit that and then re-make to add your own pre-defined section names.

module cosmosis_section_names


    implicit none
{0}

end module

"""

c_template = """
// This header file is auto-generated from the file cosmosis/datablock/section_names.txt.
// Edit that and then re-make to add your own pre-defined section names.

{0}

"""

python_template = """
# This module is auto-generated from the file cosmosis/datablock/section_names.txt.
# Edit that and then re-make to add your own pre-defined section names.

{0}

"""

def generate_python(section_names, filename):
	sections = "\n".join('{0} = "{0}"'.format(name) for name in section_names)
	open(filename,'w').write(python_template.format(sections))

def generate_fortran(section_names, filename):
	sections = "\n".join('    character(*), parameter :: {0}_section = "{0}"'.format(name) for name in section_names)
	open(filename,'w').write(fortran_template.format(sections))

def generate_c(section_names, filename):
	sections = "\n".join('#define {0}_SECTION "{1}"'.format(name.upper(),name) for name in section_names)
	open(filename,'w').write(c_template.format(sections))

def generate(section_list_filename, language, output_filename):
	section_names = []
	for line in open(section_list_filename):
		line=line.strip()
		if line.startswith('#') or not line:
			continue
		line=line.split('#')[0].strip()
		section_names.append(line)

	if language == "c":
		generate_c(section_names, output_filename)
	elif language == "python":
		generate_python(section_names, output_filename)
	elif language == "fortran":
		generate_fortran(section_names, output_filename)
	else:
		print("Unknown language specified: %s" % language)
		sys.exit(2)

if __name__ == '__main__':
	# Required syntax is:
	#   $ python generate_sections.py <template-file> <target-language> <output-file>
	#
	if len(sys.argv) != 4:
		print("%s requires 4 arguments" % sys.argv[0])
		sys.exit(1)
	generate(sys.argv[1], sys.argv[2], sys.argv[3])



from .output_base import OutputBase
import numpy as np

class InMemoryOutput(OutputBase):
    _aliases = ["memory"]
    def __init__(self):
        super(InMemoryOutput,self).__init__()
        self.rows = []
        self.meta = {}
        self.final_meta = {}
        self.comments = []

    def _write_parameters(self, params):
        self.rows.append(params)

    def _write_metadata(self, key, value, comment):
        self.meta[key] = (value,comment)

    def _write_comment(self, comment):
        self.comments.append(comment)

    def _write_final(self, key, value, comment):
        self.final_meta[key] = (value,comment)

    def __getitem__(self, key_or_index):
        if isinstance(key_or_index, int):
            return self.rows[key_or_index]
        else:
            column_index = [c[0] for c in self.columns].index(key_or_index)
            return np.array([row[column_index] for row in self.rows])

    @classmethod
    def from_options(cls, options, resume=False):
        if resume:
            raise ValueError("Cannot resume from in-memory output")
        return cls()

    @classmethod
    def load_from_options(cls, options):
        raise ValueError("No output was saved from this run")



from .. import ParallelSampler
from . import fisher
from ...datablock import BlockError
import numpy as np
import scipy.linalg
from ...runtime import prior,utils
import sys

def compute_fisher_vector(p, cov=False):
    # use normalized parameters - fisherPipeline is a global
    # variable because it has to be picklable)
    try:
        x = fisherPipeline.denormalize_vector(p)
    except ValueError:
        print("Parameter vector outside limits: %r" % p)
        return None

    if not fisherPipeline.quiet:
        print("Running: ",x)

    #Run the pipeline, generating a data block
    data = fisherPipeline.run_parameters(x)

    #If the pipeline failed, return "None"
    #This might happen if the parameters stray into
    #a bad region.
    if data is None:
        return None

    #Get out the fisher vector.  Failing on this is definitely an error
    #since if the pipeline finishes it must have a fisher vector if it
    #has been acceptably designed.
    v = []
    for like_name in fisherPipeline.likelihood_names:
        v.append(data["data_vector", like_name + "_theory"])

    v = np.concatenate(v)
    #Might be only length-one, conceivably, so convert to a vector
    v = np.atleast_1d(v)

    # If we don't need the cov mat for this run just return now
    if not cov:
        return v

    # Otherwise calculate the covmat too.
    M = []
    for like_name in fisherPipeline.likelihood_names:
        M.append(data["data_vector", like_name + "_inverse_covariance"])

    M = scipy.linalg.block_diag(*M)
    M = np.atleast_2d(M)

    #Return numpy vector
    return v, M

class SingleProcessPool(object):
    def map(self, function, tasks):
        return list(map(function, tasks))

class FisherSampler(ParallelSampler):
    sampler_outputs = []
    parallel_output = False
    understands_fast_subspaces = True

    def config(self):
        #Save the pipeline as a global variable so it
        #works okay with MPI
        global fisherPipeline
        fisherPipeline = self.pipeline
        self.step_size = self.read_ini("step_size", float, 0.01)
        self.tolerance = self.read_ini("tolerance", float, 0.01)
        self.maxiter = self.read_ini("maxiter", int, 10)
        self.use_numdifftools = self.read_ini("use_numdifftools", bool, False)

        if self.output:
            for p in self.pipeline.extra_saves:
                name = '%s--%s'%p
                print("NOTE: You set extra_output to include parameter %s in the parameter file" % name)
                print("      But the Fisher Sampler cannot do that, so this will be ignored.")
                self.output.del_column(name)

        self.converged = False

    def compute_prior_matrix(self):
        #We include the priors as an additional matrix term
        #This is just added to the fisher matrix
        n = len(self.pipeline.varied_params)
        P = np.zeros((n,n))
        for i, param in enumerate(self.pipeline.varied_params):
            if isinstance(param.prior, prior.GaussianPrior) or isinstance(param.prior, prior.TruncatedGaussianPrior):
                print("Applying additional prior sigma = {0} to {1}".format(param.prior.sigma, param))
                print("This will be assumed to be centered at the parameter center regardless of what the ini file says")
                print("The limits of the parameter will also not be respected.") 
                print()
                P[i,i] = 1./param.prior.sigma**2
            elif isinstance(param.prior, prior.ExponentialPrior) or isinstance(param.prior, prior.TruncatedExponentialPrior):
                print("There is an exponential prior applied to parameter {0}".format(param))
                print("This is *not* accounted for in the Fisher matrix")
                print()
            #uniform prior should have no effect on the fisher matrix.
            #at least up until the assumptions of the FM are violated anyway
        return P




    def execute(self):
        #Load the starting point and covariance matrix
        #in the normalized space, either from the values
        #file or a previous sampler
        start_vector = self.start_estimate()

        if len(self.pipeline.varied_params)==0:
            raise ValueError("Your values file did not include any varied parameters so we cannot make a Fisher matrix")

        for i,x in enumerate(start_vector):
            self.output.metadata("mu_{0}".format(i), x)
        start_vector = self.pipeline.normalize_vector(start_vector)

        #calculate the fisher matrix.
        #right now just a single step
        if self.use_numdifftools:
            fisher_class = fisher.NumDiffToolsFisher
        else:
            fisher_class = fisher.Fisher
        fisher_calc = fisher_class(compute_fisher_vector, start_vector, 
            self.step_size, self.tolerance, self.maxiter, pool=self.pool)

        try:
            fisher_matrix = fisher_calc.compute_fisher_matrix()
        except fisher.FisherParameterError as error:
            param = str(self.pipeline.varied_params[error.parameter_index])
            if error.parameter_index==0:
                message = """
There was an error running the pipeline for the Fisher Matrix for parameter:
{}
Since this is the first parameter this might indicate a general error in the pipeline.
You might want to check with the "test" sampler.

It might also indicate that the parameter lower or upper limit is too close to its
starting value so the points used to calculate the derivative are outside the range.
If that is the case you should try calculating the Fisher Matrix at a different starting point.
""".format(param)
            else:
                message = """
There was an error running the pipeline for the Fisher Matrix for parameter:
{}

This probably indicates that the parameter lower or upper limit is too close to its
starting value, so the points used to calculate the derivative are outside the range.
If that is the case you should try calculating the Fisher Matrix at a different starting point.
""".format(param)
            raise ValueError(message)

        fisher_matrix = self.pipeline.denormalize_matrix(fisher_matrix,inverse=True)

        P = self.compute_prior_matrix()
        fisher_matrix += P

        self.converged = True

        if self.converged:
            for row in fisher_matrix:
                self.output.parameters(row)
        try:
            covariance_matrix = utils.symmetric_positive_definite_inverse(fisher_matrix)
            self.distribution_hints.set_cov(covariance_matrix)
        except ValueError:
            sys.stderr.write("Generated covariance matrix was not positive definite - beware! ")

    def is_converged(self):
        return self.converged




# This module is auto-generated from the file cosmosis/datablock/section_names.txt.
# Edit that and then re-make to add your own pre-defined section names.

likelihoods = "likelihoods"
data_vector = "data_vector"
cosmological_parameters = "cosmological_parameters"
halo_model_parameters = "halo_model_parameters"
baryon_parameters = "baryon_parameters"
shear_calibration_parameters = "shear_calibration_parameters"
number_density_params = "number_density_params"
post_friedmann_parameters = "post_friedmann_parameters"
intrinsic_alignment_parameters = "intrinsic_alignment_parameters"
intrinsic_power = "intrinsic_power"
galaxy_intrinsic_power = "galaxy_intrinsic_power"
matter_intrinsic_power = "matter_intrinsic_power"
wl_number_density = "wl_number_density"
matter_power_nl = "matter_power_nl"
matter_power_lin = "matter_power_lin"
matter_power_lin_cdm_baryon = "matter_power_lin_cdm_baryon"
matter_galaxy_power = "matter_galaxy_power"
galaxy_power = "galaxy_power"
shear_xi = "shear_xi"
shear_cl = "shear_cl"
shear_cl_gg = "shear_cl_gg"
shear_cl_gi = "shear_cl_gi"
shear_cl_ii = "shear_cl_ii"
galaxy_cl = "galaxy_cl"
cmb_cl = "cmb_cl"
lss_autocorrelation = "lss_autocorrelation"
gal_matter_power_lin = "gal_matter_power_lin"
linear_cdm_transfer = "linear_cdm_transfer"
sigma_r_lin = "sigma_r_lin"
sigma_r_nl = "sigma_r_nl"
galaxy_xi = "galaxy_xi"
galaxy_shear_xi = "galaxy_shear_xi"
galaxy_intrinsic_xi = "galaxy_intrinsic_xi"
planck = "planck"
intrinsic_alignment_field = "intrinsic_alignment_field"
fgas_parameters = "fgas_parameters"
shear_calibration = "shear_calibration"
supernova_params = "supernova_params"
bias_field = "bias_field"
galaxy_luminosity_function = "galaxy_luminosity_function"
distances = "distances"
de_equation_of_state = "de_equation_of_state"
growth_parameters = "growth_parameters"
test_parameters = "test_parameters"
mass_function = "mass_function"
evs = "evs"
clusters = "clusters"




from setuptools import setup, find_packages
from setuptools.command.install import install
from distutils.command.build import build
from distutils.command.clean import clean
from distutils.command.sdist import sdist
import pkg_resources
import subprocess

import os
import sys

version = open('cosmosis/version.py').read().split('=')[1].strip().strip("'")


f90_mods = [
    "datablock/cosmosis_section_names.mod",
    "datablock/cosmosis_types.mod",
    "datablock/cosmosis_wrappers.mod",
    "datablock/cosmosis_modules.mod",
]

scripts = [
    'bin/cosmosis',
    'bin/cosmosis-configure',
    'bin/cosmosis-extract',
    'bin/cosmosis-sample-fisher',
    'bin/cosmosis-postprocess',
]

c_headers = [
    "datablock/c_datablock.h",
    "datablock/datablock_logging.h",
    "datablock/datablock_types.h",
    "datablock/cosmosis_constants.h",
    "datablock/datablock_status.h",
    "datablock/section_names.h",
]

cc_headers = [
    "datablock/clamp.hh",
    "datablock/entry.hh",
    "datablock/fakearray.hh",
    "datablock/ndarray.hh",
    "datablock/datablock.hh",
    "datablock/exceptions.hh",
    "datablock/mdarraygen.hh",
    "datablock/section.hh"
]

datablock_libs = ["datablock/libcosmosis.so"]

sampler_libs = ["samplers/multinest/multinest_src/libnest3.so",
                "samplers/multinest/multinest_src/libnest3_mpi.so",
                "samplers/polychord/polychord_src/libchord.so",
                "samplers/polychord/polychord_src/libchord_mpi.so",
                "samplers/minuit/minuit_wrapper.so"]


testing_files = [
    "test/libtest/c_datablock_complex_array_test.c",
    "test/libtest/c_datablock_double_array_test.c",
    "test/libtest/c_datablock_int_array_test.c",
    "test/libtest/c_datablock_multidim_complex_array_test.c",
    "test/libtest/c_datablock_multidim_double_array_test.c",
    "test/libtest/c_datablock_multidim_int_array_test.c",
    "test/libtest/c_datablock_test.c",
    "test/libtest/cosmosis_test.F90",
    "test/libtest/cosmosis_tests.supp",
    "test/libtest/datablock_test.cc",
    "test/libtest/entry_test.cc",
    "test/libtest/ndarray_test.cc",
    "test/libtest/section_test.cc",
    "test/libtest/test_c_datablock_scalars.h",
    "test/libtest/test_c_datablock_scalars.template",
    "test/libtest/Makefile",
]

other_files = ["postprocessing/latex.ini"]

compilers_config = ["config/compilers.mk", "config/subdirs.mk"]

# if sys.platform == 'darwin':
#     from distutils import sysconfig
#     vars = sysconfig.get_config_vars()
#     vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')

def get_COSMOSIS_SRC_DIR():
    cosmosis_src_dir = os.path.join(os.getcwd(), "cosmosis")
    return cosmosis_src_dir

def compile_library():
    cosmosis_src_dir = get_COSMOSIS_SRC_DIR()
    env = os.environ.copy()
    env["COSMOSIS_SRC_DIR"] = cosmosis_src_dir

    # If we are in a conda build env then the appropriate
    # variable is called PREFIX. Otherwise if we are on
    # a user's build env then it is called CONDA_PREFIX
    if os.environ.get("CONDA_BUILD", "0") == "1":
        prefix_name = "PREFIX"
        conda_build = True
    else:
        prefix_name = "CONDA_PREFIX"
        conda_build = False

    conda = env.get(prefix_name)
    if conda:
        # User can switch on COSMOSIS_OMP manually, but it should
        # always be on for conda.
        env["COSMOSIS_OMP"] = "1"

        # We also need Lapack to build some of the samplers
        env["LAPACK_LINK"] = f"-L{conda}/lib -llapack"
        # and minuit for that sampler
        env['MINUIT2_LIB'] = f"{conda}/lib"
        env['MINUIT2_INC'] = f"{conda}/include/Minuit2"

        if conda_build:
            env["COSMOSIS_ALT_COMPILERS"] = "1"
            env['USER_FFLAGS'] = env['FFLAGS']
            env['USER_CFLAGS'] = env['CFLAGS']
            env['USER_CXXFLAGS'] = env['CXXFLAGS']

        # and the MPI compiler
        env["MPIFC"] = "mpif90"

    env['FC'] = env.get('FC', 'gfortran')

    subprocess.check_call(["make"], env=env, cwd="cosmosis")
    

def clean_library():
    cosmosis_src_dir = get_COSMOSIS_SRC_DIR()
    env = {"COSMOSIS_SRC_DIR": cosmosis_src_dir,}
    subprocess.check_call(["make", "clean"], env=env, cwd="cosmosis")


class my_build(build):
    def run(self):
        compile_library()
        super().run()

class my_sdist(sdist):
    def run(self):
        self.distribution.package_data['cosmosis'] = [
            d for d in self.distribution.package_data['cosmosis']
            if d not in sampler_libs + datablock_libs + f90_mods]
        super().run()

class my_install(install):
    def __init__(self, dist):
        install.__init__(self, dist)
        self.build_args = {}
        if self.record is None:
            self.record = "install-record.txt"

    def run(self):
        super().run()

class my_clean(clean):
    def run(self):
        clean_library()
        super().run()

requirements = [
    "pyyaml",
    "future",
    "emcee",
    "numpy",
    "scipy",
    "matplotlib",
    "pybind11",
    "pyyaml",
    "scipy",
    "threadpoolctl",
    "emcee",
    "dynesty",
    "zeus-mcmc",
]

all_package_files = (datablock_libs + sampler_libs
                            + c_headers + cc_headers + f90_mods 
                            + compilers_config + testing_files + other_files)


setup(name = 'cosmosis',
        description       = "A testbed stand-alone installation of the CosmoSIS project. Not ready for primetime!",
        author            = "Joe Zuntz",
        author_email      = "joezuntz@googlemail.com",
        url               = "https://bitbucket.org/joezuntz/cosmosis",  
        packages = find_packages(),
        package_data = {"cosmosis" : all_package_files},
        scripts = scripts,
        install_requires = requirements,
        cmdclass={"install"   : my_install,
                "build"     : my_build,
                "build_ext" : my_build,
                "clean"     : my_clean,
                "sdist"     : my_sdist

                },
        version=version,
        )




from .output_base import OutputBase
from . import utils
from ..runtime.utils import mkdir
import numpy as np
import os
from glob import glob
from collections import OrderedDict

comment_indicator = "_cosmosis_comment_indicator_"


class TextColumnOutput(OutputBase):
    FILE_EXTENSION = ".txt"
    _aliases = ["text", "txt"]

    def __init__(self, filename, rank=0, nchain=1, delimiter='\t', lock=True, resume=False):
        super(TextColumnOutput, self).__init__()
        self.delimiter = delimiter

        #If filename already ends in .txt then remove it for a moment
        if filename.endswith(self.FILE_EXTENSION):
            filename = filename[:-len(self.FILE_EXTENSION)]

        if nchain > 1:
            filename = filename + "_{}".format(rank+1)

        self.filename_base = filename

        self._filename = filename + self.FILE_EXTENSION

        dirname, _ = os.path.split(self._filename)
        mkdir(dirname)

        if resume and utils.file_exists_and_is_empty(self._filename):
            print("You set resume=T but the file {} is empty so I will start afresh".format(self._filename))
            self._file = open(self._filename, "w")
            self.resumed = False
        elif resume and os.path.exists(self._filename):
            print("Note: You set resume=T so I will resume from file {}".format(self._filename))
            self._file = open(self._filename, "r+")
            # Jump to the end of the file
            self._file.seek(0,2)
            self.resumed = True
        else:
            if resume:
                print("Note: You set resume=T but the file {} does not exist so I will start a new one".format(self._filename))
            self._file = open(self._filename, "w")
            self.resumed = False
        if lock:
            try:
                self.lock_file(self._file)
            except IOError:
                error_msg = """
Another CosmoSIS process was trying to use the same output file (%s). 
This means one of three things:
1) you were trying to use MPI but left out the --mpi flag
2) you have another CosmoSIS run going trying to use the same filename
3) your file system cannot cope with file locks properly.  
In the last case you can set lock=F in the [output] section to disable this feature.
""" % self._filename
                raise IOError(error_msg)

        #also used to store comments:
        self._metadata = OrderedDict()
        self._final_metadata = OrderedDict()

        # For resetting
        self._start_mark = None

    def _close(self):
        self._flush_metadata(self._final_metadata)
        self._final_metadata={}
        self._file.close()

    def _flush_metadata(self, metadata):
        for (key,(value,comment)) in list(metadata.items()):
            if key.startswith(comment_indicator):
                self._file.write("## %s\n"%value.strip())
            elif comment:
                self._file.write('#{k}={v} #{c}\n'.format(k=key,v=value,c=comment))
            else:
                self._file.write('#{k}={v}\n'.format(k=key,v=value,c=comment))


    def _begun_sampling(self, params):
        #write the name line
        if not self.resumed:
            name_line = '#'+self.delimiter.join(c[0] for c in self.columns) + '\n'
            self._file.write(name_line)
            #now write any metadata.
            #text mode does not support comments
            self._flush_metadata(self._metadata)
        self._metadata={}
        self._start_mark = self._file.tell()

    def _write_metadata(self, key, value, comment=''):
        #We save the metadata until we get the first 
        #parameters since up till then the columns can
        #be changed
        #In the text mode we cannot write more metadata
        #after sampling has begun (because it goes at the top).
        #What should we do?
        self._metadata[key]= (value, comment)

    def _write_comment(self, comment):
        #save comments along with the metadata - nice as 
        #preserves order
        # remove line breaks in comments
        comment = comment.replace("\n", " ")
        self._metadata[comment_indicator +
                       "_%d" % (len(self._metadata))] = (comment,None)

    def _write_parameters(self, params):
        line = self.delimiter.join(str(x) for x in params) + '\n'
        self._file.write(line)

    def _write_final(self, key, value, comment=''):
        #I suppose we can put this at the end - why not?
        self._final_metadata[key]= (value, comment)

    def _flush(self):
        self._file.flush()

    def reset_to_chain_start(self):
        # On the first iteration the start mark is not set until we call
        # output the first time.
        if self._start_mark is None:
            return
        self._file.seek(self._start_mark)
        self._file.truncate()
        self._file.flush()

    def name_for_sampler_resume_info(self):
        return self.filename_base + '.sampler_status'



    @classmethod
    def from_options(cls, options, resume=False):
        #look something up required parameters in the ini file.
        #how this looks will depend on the ini 
        filename = options['filename']
        delimiter = options.get('delimiter', '\t')
        rank = options.get('rank', 0)
        nchain = options.get('parallel', 1)
        lock = utils.boolean_string(options.get('lock', True))
        return cls(filename, rank, nchain, delimiter=delimiter, lock=lock, resume=resume)

    @classmethod
    def load_from_options(cls, options):
        filename = options['filename']
        delimiter = options.get('delimiter', None)

        cut = False
        if filename.endswith(cls.FILE_EXTENSION):
            filename = filename[:-len(cls.FILE_EXTENSION)]
            cut = True
        # first look for serial file
        if os.path.exists(filename+cls.FILE_EXTENSION):
            datafiles = [filename+cls.FILE_EXTENSION]
        elif os.path.exists(filename) and not cut:
            datafiles = [filename]
        else:
            datafiles = glob(filename+"_[0-9]*"+cls.FILE_EXTENSION)
            if not datafiles:
                raise RuntimeError("No datafiles found starting with %s!"%filename)

        #Read the metadata
        started_data = False
        metadata = []
        final_metadata = []
        data = []
        comments = []
        column_names = None

        for datafile in datafiles:
            print('LOADING CHAIN FROM FILE: ', datafile)
            chain = []
            chain_metadata = {}
            chain_final_metadata = {}
            chain_comments = []
            for i,line in enumerate(open(datafile)):
                line = line.strip()
                if not line: continue
                if line.startswith('#'):
                    #remove the first #
                    #if there is another then this is a comment,
                    #not metadata
                    line=line[1:]
                    if i == 0:
                        column_names = line.split()
                    elif line.startswith('#'):
                        comment = line[1:]
                        chain_comments.append(comment)
                    else:
                        #parse form '#key=value #comment'
                        if line.count('#') == 0:
                            key_val = line.strip()
                            comment = ''
                        else:
                            key_val, comment = line.split('#', 1)
                        key,val = key_val.split('=',1)
                        val = utils.parse_value(val)
                        if started_data:
                            chain_final_metadata[key] = val
                        else:
                            chain_metadata[key] = val
                else:
                    started_data = True
                    words = line.split(delimiter)
                    vals = [float(word) for word in words]
                    chain.append(vals)
            ncol = len(column_names)
            line_lengths = np.array([len(row) for row in chain])
            #strip off the last line if it is incompletely written as often
            #the chain is interrupted
            if line_lengths[-1]!=ncol:
                print("Skipping last line of chain as it seems to have been cut off")
                print("This could conceivably cause problems for some samplers, though")
                print("not the ones like metropolis and emcee where it is most likely to happen.")
                print("If any more lines have the wrong length then this will raise an error.")
                print()
                print("You should probably check the final lines of the other files for errors")
                print("that are harder to detect, like values being truncated.")
                print()
                chain = chain[:-1]
                line_lengths = line_lengths[:-1]
            #if any more are the wrong length then something has gone wrong:
            if np.any(line_lengths!=ncol):
                raise ValueError("Your chain file is corrupted somehow: not all the lines have {} columns".format(ncol))


            data.append(np.array(chain))
            metadata.append(chain_metadata)
            final_metadata.append(chain_final_metadata)
            comments.append(chain_comments)

        if column_names is None:
            raise ValueError("Could not find column names header in file starting %s"%filename)

        return column_names, data, metadata, comments, final_metadata



import scipy.stats
import numpy as np
from functools import wraps
from distutils.version import LooseVersion


def not_with_weights(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        self=args[0]
        if self.weights is not None:
            raise ValueError("The standard KDE functions have not been updated to use weights on points")
    return wrapper


# Newer versions of scipy have the weights feature built in automatically,
# and the cosmosis version of that feature clashes with it.
# So check which version we are using and define KDE differently in each case.
if LooseVersion(scipy.__version__) > LooseVersion('1.1.0'):
    class KDE(scipy.stats.kde.gaussian_kde):
        def __init__(self, points, factor=1.0, weights=None):
            points = np.array(np.atleast_2d(points))

            self._factor = factor
            d, n = points.shape
            self.norms = []
            normalized_points = []

            for column in points:
                col_mean = column.mean()
                col_std = column.std()
                self.norms.append((col_mean,col_std))
                normalized_points.append((column-col_mean)/col_std)
            super(KDE,self).__init__(normalized_points, weights=weights)

        def covariance_factor(self):
            return self.scotts_factor() * self._factor

        def grid_evaluate(self, n, ranges):
            if isinstance(ranges,tuple):
                ranges = [ranges]
            slices = [slice(xmin,xmax,n*1j) for (xmin,xmax) in ranges]
            grids = np.mgrid[slices]
            axes = [ax.squeeze() for ax in np.ogrid[slices]]
            flats = [(grid.flatten()-norm[0])/norm[1]
                     for (grid,norm) in zip(grids,self.norms)]

            shape = grids[0].shape
            flats = np.array(flats)
            like_flat = self.evaluate(flats)
            like = like_flat.reshape(*shape)
            if len(axes)==1:
                axes = axes[0]
            return axes,like

        def normalize_and_evaluate(self, points):
            points = np.array([(p-norm[0])/norm[1] for norm, p in zip(self.norms, points)])
            return self.evaluate(points)


# earlier scipy version
else:
    class KDE(scipy.stats.kde.gaussian_kde):
        def __init__(self, points, factor=1.0, weights=None):
            points = np.array(np.atleast_2d(points))

            self._factor = factor
            d, n = points.shape
            self.norms = []
            self.weights=weights
            normalized_points = []
            #assert d<=2, "KDE is for 1D 2D only"
            for column in points:
                col_mean = column.mean()
                col_std = column.std()
                self.norms.append((col_mean,col_std))
                normalized_points.append((column-col_mean)/col_std)
            super(KDE,self).__init__(normalized_points)

        def covariance_factor(self):
            return self.scotts_factor() * self._factor

        def grid_evaluate(self, n, ranges):
            if isinstance(ranges,tuple):
                ranges = [ranges]
            slices = [slice(xmin,xmax,n*1j) for (xmin,xmax) in ranges]
            grids = np.mgrid[slices]
            axes = [ax.squeeze() for ax in np.ogrid[slices]]
            flats = [(grid.flatten()-norm[0])/norm[1]
                     for (grid,norm) in zip(grids,self.norms)]

            shape = grids[0].shape
            flats = np.array(flats)
            like_flat = self.evaluate(flats)
            like = like_flat.reshape(*shape)
            if len(axes)==1:
                axes = axes[0]
            return axes,like

        def normalize_and_evaluate(self, points):
            points = np.array([(p-norm[0])/norm[1] for norm, p in zip(self.norms, points)])
            return self.evaluate(points)

        def evaluate(self, points):
            """Evaluate the estimated pdf on a set of points.

            JAZ: Overriding this function to allow the use of weights.

            Parameters
            ----------
            points : (# of dimensions, # of points)-array
                Alternatively, a (# of dimensions,) vector can be passed in and
                treated as a single point.

            Returns
            -------
            values : (# of points,)-array
                The values at each point.

            Raises
            ------
            ValueError : if the dimensionality of the input points is different than
                         the dimensionality of the KDE.

            """
            points = np.atleast_2d(points)
            d, m = points.shape
            if d != self.d:
                if d == 1 and m == self.d:
                    # points was passed in as a row vector
                    points = np.reshape(points, (self.d, 1))
                    m = 1
                else:
                    msg = "points have dimension %s, dataset has dimension %s" % (d,
                        self.d)
                    raise ValueError(msg)

            result = np.zeros((m,), dtype=np.float)


            if m >= self.n:
                weights = self.weights
                if weights is None: weights=np.repeat(1.0,self.n)
                # there are more points than data, so loop over data
                for i in range(self.n):
                    diff = self.dataset[:, i, np.newaxis] - points
                    tdiff = np.dot(self.inv_cov, diff)
                    energy = np.sum(diff*tdiff,axis=0) / 2.0
                    result = result + weights[i]*np.exp(-energy)
            else:
                weights = self.weights
                if weights is None: weights=1

                # loop over points
                for i in range(m):
                    diff = self.dataset - points[:, i, np.newaxis]
                    tdiff = np.dot(self.inv_cov, diff)
                    energy = np.sum(diff * tdiff, axis=0) / 2.0
                    result[i] = np.sum(weights*np.exp(-energy), axis=0)/np.sum(weights)

            result = result / self._norm_factor

            return result

        @not_with_weights
        def integrate_gaussian(self, mean, cov):
            return super(KDE,self).integrate_gaussian(mean, cov)

        @not_with_weights
        def integrate_box_1d(self, low, high):
            return super(KDE,self).integrate_box_1d(low, high)

        @not_with_weights
        def integrate_box(self, low_bounds, high_bounds, maxpts=None):
            return super(KDE,self).integrate_box(low_bounds, high_bounds, maxpts=maxpts)

        @not_with_weights
        def integrate_kde(self, other):
            return super(KDE, self).integrate_kde(other)

        @not_with_weights
        def resample(self):
            return super(KDE, self).resample()

        def _compute_covariance(self):
            """Computes the covariance matrix for each Gaussian kernel using
            covariance_factor().

            JAZ overridden to use weights
            """
            self.factor = self.covariance_factor()
            # Cache covariance and inverse covariance of the data
            if not hasattr(self, '_data_inv_cov'):
                #JAZ this is the overridden bit
                self._data_covariance = np.atleast_2d(np.cov(self.dataset, aweights=self.weights, bias=False))
                self._data_inv_cov = np.linalg.inv(self._data_covariance)

            self.covariance = self._data_covariance * self.factor**2
            self.inv_cov = self._data_inv_cov / self.factor**2
            self._norm_factor = np.sqrt(np.linalg.det(2*np.pi*self.covariance)) * self.n



import numpy as np

def std_weight(x, w):
    mu = mean_weight(x,w)
    r = x-mu
    return np.sqrt((w*r**2).sum() / w.sum())

def mean_weight(x, w):
    return (x*w).sum() / w.sum()

def median_weight(x, w):
	a = np.argsort(x)
	w = w[a]
	x = x[a]
	wc = np.cumsum(w)
	wc/=wc[-1]
	return np.interp(0.5, wc, x)

def percentile_weight(x, w, p):
	a = np.argsort(x)
	w = w[a]
	x = x[a]
	wc = np.cumsum(w)
	wc/=wc[-1]
	return np.interp(p/100., wc, x)



def find_asymmetric_errorbars(levels, v, weights=None):
    from ..plotting.kde import KDE
    import scipy.optimize
    N = len(v)

    #Generate and normalize weights
    if weights is None:
        weights = np.ones(N)
    weights = weights / weights.sum()

    #Normalize the parameter values
    mu = mean_weight(v,weights)
    sigma = std_weight(v,weights)
    x = (v-mu)/sigma

    #Build the P(x) estimator
    K=KDE(x, weights=weights)

    #Generate the axis over which get P(x)
    xmin = x[weights>0].min()
    xmax = x[weights>0].max()
    X = np.linspace(xmin,xmax,500)
    Y = K.normalize_and_evaluate(np.atleast_2d(X))
    Y/=Y.max()

    peak1d = X[Y.argmax()]
    peak1d = peak1d*sigma+mu

    #Take the log but suppress the log(0) warning
    old_settings = np.geterr()
    np.seterr(all='ignore')
    Y=np.log(Y)
    np.seterr(**old_settings)  # reset to default


    #Calculate the levels
    def objective(level, target_weight):
        w = np.where(Y>level)[0]
        if len(w) == 0:
            weight_inside = 0.0
        else:
            low = X[w].min()
            high = X[w].max()
            inside = (x>=low) & (x<=high)
            weight_inside = weights[inside].sum()
        return (weight_inside - target_weight)

    limits = []
    for target_weight in levels:
        level = scipy.optimize.bisect(objective, Y[np.isfinite(Y)].min(), Y.max(), args=(target_weight,))
        w = np.where(Y>level)[0]
        low = X[w].min()
        high = X[w].max()
        #Convert back to origainal space
        low = low*sigma+mu
        high = high*sigma+mu
        limits.append((low,high))

    return peak1d, limits



from . import plots
from . import statistics
from .postprocess_base import PostProcessor, postprocessor_registry
import numpy as np
import collections
import hashlib

def postprocessor_for_sampler(sampler):
	return postprocessor_registry.get(sampler)

def blinding_value(name, seed):
    #hex number derived from code phrase
    m = hashlib.md5(name).hexdigest()
    #convert to decimal
    s = int(m, 16) + seed
    # last 8 digits
    f = s%100000000
    # turn 8 digit number into value between 0 and 1
    g = f*1e-8
    #get value between -1 and 1
    return g*2-1


def additive_blinding(postprocessors, seed):
	factors = collections.defaultdict(lambda: -np.inf)
	for P in postprocessors:
		for c,col in enumerate(P.colnames):
			col = col.lower()
			#Skip likelihood columns
			if col in ['like','post', 'weight', 'log_weight', 'old_weight', 'old_log_weight']:continue
			#Work out the approximate scale of the parameters.
			f = P.approximate_scale_ceiling(c)
			#We use the max approx scale over all the chains
			factors[col] = max(factors[col], f)

	#print out scale info
	for col,f in list(factors.items()):
		print("Blinding additive value for %s ~ %.1e" % (col, f))

	for P in postprocessors:
		for c,col in enumerate(P.colnames):
			col = col.lower()
			if col in ['like','post', 'weight', 'log_weight', 'old_weight', 'old_log_weight']:continue
			#add +- the ceiling of the typical standard deviation. e.g. if width is ~ 20% add up to 3.0
			b = 3 * factors[col] * blinding_value(col,seed)
			P.additive_blind_column(c,b)

def multiplicative_blinding(postprocessors, seed):
	#print out scale info
	scale = 0.5
	print("Blinding all parameters by -50% to +50%")
	for P in postprocessors:
		for c,col in enumerate(P.colnames):
			col = col.lower()
			#blind by up to 50%
			f = blinding_value(col,seed) * 0.5
			P.multiplicative_blind_column(c,f)




class MetropolisHastingsProcessor(PostProcessor):
	elements=[
		plots.MetropolisHastingsPlots1D,
		plots.MetropolisHastingsPlots2D,
		statistics.MetropolisHastingsStatistics,
		statistics.MetropolisHastingsCovariance,
		statistics.Citations,
	]
	def reduced_col(self, name, stacked=True):
		cols = self.get_col(name, stacked=False)
		burn = self.options.get("burn", 0)
		thin = self.options.get("thin", 1)

		if 0.0<burn<1.0:
			burn = int(len(cols[0])*burn)
		else:
			burn = int(burn)
		cols = [col[burn::thin] for col in cols]
		if stacked:
			return np.concatenate(cols)
		else:
			return cols

	def posterior_sample(self):
		"""
		A posterior sample of MCMC is just all the samples.

		Return an array of Trues with the same length as the chain

		"""
		n = self.reduced_col(self.colnames[0]).size
		return np.ones(n, dtype=bool)



class EmceeProcessor(MetropolisHastingsProcessor):
	sampler="emcee"
	def reduced_col(self, name, stacked=True):
		cols = self.get_col(name, stacked=False)
		burn = self.options.get("burn", 0)
		thin = self.options.get("thin", 1)

		if 0.0<burn<1.0:
			burn = int(len(cols[0])*burn)
		else:
			burn = int(burn)

		cols = [col[burn::] for col in cols]

		if thin!=1:
			walkers=self.sampler_option("walkers")
			index = np.arange(len(cols[0]), dtype=np.int64)
			index = index//int(walkers)
			w = (index%thin)==0
			cols = [col[w] for col in cols]

		if stacked:
			return np.concatenate(cols)
		else:
			return cols

class ZeusProcessor(EmceeProcessor):
	sampler="zeus"

class KombineProcessor(MetropolisHastingsProcessor):
	sampler="kombine"
	pass


class PymcProcessor(MetropolisHastingsProcessor):
	sampler="pymc"
	pass

class MetropolisProcessor(MetropolisHastingsProcessor):
	sampler="metropolis"
	elements=[
		plots.MetropolisHastingsPlots1D,
		plots.MetropolisHastingsPlots2D,
		statistics.MetropolisHastingsStatistics,
		statistics.MetropolisHastingsCovariance,
		statistics.DunkleyTest,
		statistics.GelmanRubinStatistic,
		statistics.Citations,
		]


class WeightedMetropolisProcessor(MetropolisHastingsProcessor):
	sampler="weighted_metropolis"
	elements=[
		plots.WeightedMetropolisPlots1D,
		plots.WeightedMetropolisPlots2D,
		statistics.WeightedMetropolisStatistics,
		statistics.WeightedMetropolisHastingsCovariance,
		#statistics.DunkleyTest,
		statistics.Citations,		
		]

	def weight_col(self):
		if hasattr(self, "_weight_col"):
			return self._weight_col
		if self.has_col("weight"):
			w = MetropolisHastingsProcessor.reduced_col(self, "weight").copy()
			logw = np.log(w)
		elif self.has_col("log_weight"):
			logw = MetropolisHastingsProcessor.reduced_col(self, "log_weight").copy()
		else:
			raise ValueError("No 'weight' or 'log_weight' column found in chain.")


		if self.has_col("old_weight"):
			old_w = MetropolisHastingsProcessor.reduced_col(self, "old_weight").copy()
			old_logw = np.log(old_w)
			logw += old_logw
			print("Including old_weight in weight")
		elif self.has_col("old_log_weight"):
			old_logw = MetropolisHastingsProcessor.reduced_col(self, "old_log_weight").copy()
			logw += old_logw
			print("Including old_log_weight in weight")
		logw-=np.nanmax(logw)
		self._weight_col = np.exp(logw)
		return self._weight_col    

	def posterior_sample(self):
		"""
		Weighted chains are *not* drawn from the posterior distribution
		but we do have the information we need to construct such a sample.

		This function returns a boolean array with True where we should
		use the sample at that index, and False where we should not.

		"""
		w = self.weight_col()
		w = w / w.max()
		u = np.random.uniform(size=w.size)
		return u<w


class ImportanceProcessor(WeightedMetropolisProcessor):
	sampler="importance"
	elements=[
		plots.WeightedMetropolisPlots1D,
		plots.WeightedMetropolisPlots2D,
		statistics.WeightedMetropolisStatistics,
		# statistics.DunkleyTest,
		statistics.Citations,		
		]

class GridProcessor(PostProcessor):
	elements=[
		plots.GridPlots1D,
		plots.GridPlots2D,
		statistics.GridStatistics,
		statistics.Citations,
		]
	sampler="grid"



class TestProcessor(PostProcessor):
	elements = [plots.TestPlots]
	sampler="test"
	cosmosis_standard_output=False

	def load(self, ini):
		if isinstance(ini, str) and os.path.isdir(ini):			
			self._sampler_options['save_dir'] = ini
		else:
			for key,val in ini.items("test"):
				self.sampler_options[key]=str(val)



class MultinestProcessor(WeightedMetropolisProcessor):
	elements = [
		plots.MultinestPlots1D, 
		plots.MultinestPlots2D,
		statistics.MultinestStatistics,
		statistics.MultinestCovariance,
		statistics.Citations,
		]
	sampler="multinest"
	def reduced_col(self, name, stacked=True):
		#we only use the last n samples from a multinest output
		#file.  And omit zero-weighted samples.
		n = int(self.final_metadata[0]["nsample"])
		col = self.get_col(name)
		w = self.get_col("weight")[-n:]
		return col[-n:][w>0]
		
	def weight_col(self):
		if hasattr(self, "_weight_col"):
			return self._weight_col
		n = int(self.final_metadata[0]["nsample"])
		w = self.get_col("weight")[-n:]
		w = w[w>0].copy()
		self._weight_col = w
		return self._weight_col

	def posterior_sample(self):
		"""
		Multinest chains are *not* drawn from the posterior distribution
		but we do have the information we need to construct such a sample.

		This function returns a boolean array with True where we should
		use the sample at that index, and False where we should not.

		"""
		w = self.weight_col()
		w = w / w.max()
		u = np.random.uniform(size=w.size)
		return u<w

class PolyChordProcessor(MultinestProcessor):
    elements = [
            plots.PolychordPlots1D, 
            plots.PolychordPlots2D,
            statistics.PolychordStatistics,
            statistics.PolychordCovariance,
            statistics.Citations,
            ]
    sampler="polychord"



class PMCPostProcessor(WeightedMetropolisProcessor):
	sampler="pmc"
	elements = [
		plots.WeightedMetropolisPlots1D,
		plots.WeightedMetropolisPlots2D,
		statistics.WeightedMetropolisStatistics,
		#statistics.WeightedMetropolisHastingsCovariance,
		#statistics.DunkleyTest,
		statistics.Citations,		
	]
	def reduced_col(self, name, stacked=True):
		#we only use the last n samples from a multinest output
		#file.  And omit zero-weighted samples.
		n = int(self.final_metadata[0]["nsample"])
		col = self.get_col(name)
		w = self.get_col("log_weight")[-n:]
		return col[-n:][np.isfinite(w)]
		
	def weight_col(self):
		if hasattr(self, "_weight_col"):
			return self._weight_col
		n = int(self.final_metadata[0]["nsample"])
		w = self.get_col("log_weight")[-n:]
		w = w[np.isfinite(w)].copy()
		self._weight_col = np.exp(w)
		return self._weight_col

	def posterior_sample(self):
		"""
		PMC chains are *not* drawn from the posterior distribution - they have weights.
		We do have the information we need to construct such a sample.

		This function returns a boolean array with True where we should
		use the sample at that index, and False where we should not.

		"""
		w = self.weight_col()
		w = w / w.max()
		u = np.random.uniform(size=w.size)
		return u<w


class SnakeProcessor(PostProcessor):
	sampler='snake'
	elements = [
		plots.GridPlots1D,
		plots.SnakePlots2D,
		statistics.Citations,		
	]

class FisherProcessor(PostProcessor):
	sampler = 'fisher'
	elements = [
		plots.CovarianceMatrixEllipse,
		plots.CovarianceMatrixGaussians,
		statistics.CovarianceMatrix1D,
		statistics.CovarianceMatrixEllipseAreas,
		statistics.FisherFigureOfMerit,
	]


class StarProcessor(PostProcessor):
	sampler='star'
	elements=[
		plots.StarPlots,
	]


